# -*- coding: utf-8 -*-
"""
GStools subpackage providing truncated power law covariance models.

.. currentmodule:: gstools.covmodel.tpl_models

The following classes and functions are provided

.. autosummary::
   TPLGaussian
   TPLExponential
   TPLStable
"""
# pylint: disable=C0103, E1101
from __future__ import print_function, division, absolute_import

import warnings
import numpy as np
from gstools.covmodel.base import CovModel
from gstools.tools.special import tplstable_cor

__all__ = ["TPLGaussian", "TPLExponential", "TPLStable"]


# Truncated power law #########################################################


class TPLGaussian(CovModel):
    r"""Truncated-Power-Law with Gaussian modes.

    Notes
    -----
    The truncated power law is given by a superposition of scale-dependent
    variograms:

    .. math::
       \gamma_{\ell_{\mathrm{low}},\ell_{\mathrm{up}}}(r) =
       \intop_{\ell_{\mathrm{low}}}^{\ell_{\mathrm{up}}}
       \gamma(r,\lambda) \frac{\rm d \lambda}{\lambda}

    with `Gaussian` modes on each scale:

    .. math::
       \gamma(r,\lambda) &=
       \sigma^2(\lambda)\cdot\left(1-
       \exp\left[- \left(\frac{r}{\lambda}\right)^{2}\right]
       \right)\\
       \sigma^2(\lambda) &= C\cdot\lambda^{2H}

    This results in:

    .. math::
       \gamma_{\ell_{\mathrm{low}},\ell_{\mathrm{up}}}(r) &=
       \sigma^2_{\ell_{\mathrm{low}},\ell_{\mathrm{up}}}\cdot\left(1-
       H \cdot
       \frac{\ell_{\mathrm{up}}^{2H} \cdot
       E_{1+H}
       \left[\left(\frac{r}{\ell_{\mathrm{up}}}\right)^{2}\right]
       - \ell_{\mathrm{low}}^{2H} \cdot
       E_{1+H}
       \left[\left(\frac{r}{\ell_{\mathrm{low}}}\right)^{2}\right]}
       {\ell_{\mathrm{up}}^{2H}-\ell_{\mathrm{low}}^{2H}}
       \right) \\
       \sigma^2_{\ell_{\mathrm{low}},\ell_{\mathrm{up}}} &=
       \frac{C\cdot\left(\ell_{\mathrm{up}}^{2H}
       -\ell_{\mathrm{low}}^{2H}\right)}{2H}

    The "length scale" of this model is equivalent by the integration range:

    .. math::
       \ell = \ell_{\mathrm{up}} -\ell_{\mathrm{low}}

    If you want to define an upper scale truncation, you should set ``len_low``
    and ``len_scale`` accordingly.

    The following Parameters occure:

        * :math:`C>0` : The scaling factor from the Power-Law.
          This parameter will be calculated internally by the given variance.
          You can access C directly by ``model.var_raw``
        * :math:`0<H<1` : The hurst coefficient (``model.hurst``)
        * :math:`\ell_{\mathrm{low}}\geq 0` : The lower length scale truncation
          of the model (``model.len_low``)
        * :math:`\ell_{\mathrm{up}}\geq 0` : The upper length scale truncation
          of the model (``model.len_up``)

          This will be calculated internally by:

            * ``len_up = len_low + len_scale``

          That means, that the ``len_scale`` in this model actually represents
          the integration range for the truncated power law.
        * :math:`E_s(x)` is the exponential integral.

    Other Parameters
    ----------------
    **opt_arg
        The following parameters are covered by these keyword arguments
    hurst : :class:`float`, optional
        Hurst coefficient of the power law.
        Standard range: ``(0, 1)``.
        Default: ``0.5``
    len_low : :class:`float`, optional
        The lower length scale truncation of the model.
        Standard range: ``[0, 1000]``.
        Default: ``0.0``
    """

    @property
    def len_up(self):
        """:class:`float`: Upper length scale truncation of the model.

        * ``len_up = len_low + len_scale``
        """
        return self.len_low + self.len_scale

    def var_factor(self):
        r"""Factor for C (Power-Law factor) to result in variance.

        This is used to result in the right variance, which is depending
        on the hurst coefficient and the length-scale extents

        .. math::
           \frac{\ell_{\mathrm{up}}^{2H} - \ell_{\mathrm{low}}^{2H}}{2H}

        Returns
        -------
        :class:`float`
            factor
        """
        return (
            self.len_up ** (2 * self.hurst) - self.len_low ** (2 * self.hurst)
        ) / (2 * self.hurst)

    def default_opt_arg(self):
        """Defaults for the optional arguments.

            * ``{"hurst": 0.5, "len_low": 0.0}``

        Returns
        -------
        :class:`dict`
            Defaults for optional arguments
        """
        return {"hurst": 0.5, "len_low": 0.0}

    def default_arg_bounds(self):
        """Provide default boundaries for arguments.

        Given as a dictionary.
        """
        res = {
            "var": (0.0, 10000.0, "oc"),
            "len_scale": (0.0, 1000.0, "oo"),
            "nugget": (0.0, 100.0, "cc"),
        }
        return res

    def default_opt_arg_bounds(self):
        """Defaults for boundaries of the optional arguments.

            * ``{"hurst": [0, 1, "oo"], "len_low": [0, 1000, "cc"]}``

        Returns
        -------
        :class:`dict`
            Boundaries for optional arguments
        """
        return {"hurst": [0.1, 1, "oo"], "len_low": [0, 1000, "cc"]}

    def correlation(self, r):
        r"""Truncated-Power-Law with Gaussian modes - correlation function.

        If ``len_low=0`` we have a simple representation:

        .. math::
           \mathrm{cor}(r) =
           H \cdot
           E_{1+H}
           \left[
           \left(\frac{r}{\ell}\right)^{2}
           \right]

        The general case:

        .. math::
           \mathrm{cor}(r) =
           H \cdot
           \frac{\ell_{\mathrm{up}}^{2H} \cdot
           E_{1+H}
           \left[\left(\frac{r}{\ell_{\mathrm{up}}}\right)^{2}\right]
           - \ell_{\mathrm{low}}^{2H} \cdot
           E_{1+H}
           \left[\left(\frac{r}{\ell_{\mathrm{low}}}\right)^{2}\right]}
           {\ell_{\mathrm{up}}^{2H}-\ell_{\mathrm{low}}^{2H}}
        """
        # if lower limit is 0 we use the simplified version (faster)
        if np.isclose(self.len_low, 0.0):
            return tplstable_cor(r, self.len_scale, self.hurst, 2)
        return (
            self.len_up ** (2 * self.hurst)
            * tplstable_cor(r, self.len_up, self.hurst, 2)
            - self.len_low ** (2 * self.hurst)
            * tplstable_cor(r, self.len_low, self.hurst, 2)
        ) / (
            self.len_up ** (2 * self.hurst) - self.len_low ** (2 * self.hurst)
        )


class TPLExponential(CovModel):
    r"""Truncated-Power-Law with Exponential modes.

    Notes
    -----
    The truncated power law is given by a superposition of scale-dependent
    variograms:

    .. math::
       \gamma_{\ell_{\mathrm{low}},\ell_{\mathrm{up}}}(r) =
       \intop_{\ell_{\mathrm{low}}}^{\ell_{\mathrm{up}}}
       \gamma(r,\lambda) \frac{\rm d \lambda}{\lambda}

    with `Exponential` modes on each scale:

    .. math::
       \gamma(r,\lambda) &=
       \sigma^2(\lambda)\cdot\left(1-
       \exp\left[- \frac{r}{\lambda}\right]
       \right)\\
       \sigma^2(\lambda) &= C\cdot\lambda^{2H}

    This results in:

    .. math::
       \gamma_{\ell_{\mathrm{low}},\ell_{\mathrm{up}}}(r) &=
       \sigma^2_{\ell_{\mathrm{low}},\ell_{\mathrm{up}}}\cdot\left(1-
       2H \cdot
       \frac{\ell_{\mathrm{up}}^{2H} \cdot
       E_{1+2H}\left[\frac{r}{\ell_{\mathrm{up}}}\right]
       - \ell_{\mathrm{low}}^{2H} \cdot
       E_{1+2H}\left[\frac{r}{\ell_{\mathrm{low}}}\right]}
       {\ell_{\mathrm{up}}^{2H}-\ell_{\mathrm{low}}^{2H}}
       \right) \\
       \sigma^2_{\ell_{\mathrm{low}},\ell_{\mathrm{up}}} &=
       \frac{C\cdot\left(\ell_{\mathrm{up}}^{2H}
       -\ell_{\mathrm{low}}^{2H}\right)}{2H}

    The "length scale" of this model is equivalent by the integration range:

    .. math::
       \ell = \ell_{\mathrm{up}} -\ell_{\mathrm{low}}

    If you want to define an upper scale truncation, you should set ``len_low``
    and ``len_scale`` accordingly.

    The following Parameters occure:

        * :math:`C>0` : The scaling factor from the Power-Law.
          This parameter will be calculated internally by the given variance.
          You can access C directly by ``model.var_raw``
        * :math:`0<H<\frac{1}{2}` : The hurst coefficient (``model.hurst``)
        * :math:`\ell_{\mathrm{low}}\geq 0` : The lower length scale truncation
          of the model (``model.len_low``)
        * :math:`\ell_{\mathrm{up}}\geq 0` : The upper length scale truncation
          of the model (``model.len_up``)

          This will be calculated internally by:

            * ``len_up = len_low + len_scale``

          That means, that the ``len_scale`` in this model actually represents
          the integration range for the truncated power law.
        * :math:`E_s(x)` is the exponential integral.

    Other Parameters
    ----------------
    **opt_arg
        The following parameters are covered by these keyword arguments
    hurst : :class:`float`, optional
        Hurst coefficient of the power law.
        Standard range: ``(0, 1)``.
        Default: ``0.5``
    len_low : :class:`float`, optional
        The lower length scale truncation of the model.
        Standard range: ``[0, 1000]``.
        Default: ``0.0``
    """

    @property
    def len_up(self):
        """:class:`float`: Upper length scale truncation of the model.

        * ``len_up = len_low + len_scale``
        """
        return self.len_low + self.len_scale

    def var_factor(self):
        r"""Factor for C (Power-Law factor) to result in variance.

        This is used to result in the right variance, which is depending
        on the hurst coefficient and the length-scale extents

        .. math::
           \frac{\ell_{\mathrm{up}}^{2H} - \ell_{\mathrm{low}}^{2H}}{2H}

        Returns
        -------
        :class:`float`
            factor
        """
        return (
            self.len_up ** (2 * self.hurst) - self.len_low ** (2 * self.hurst)
        ) / (2 * self.hurst)

    def default_opt_arg(self):
        """Defaults for the optional arguments.

            * ``{"hurst": 0.25, "len_low": 0.0}``

        Returns
        -------
        :class:`dict`
            Defaults for optional arguments
        """
        return {"hurst": 0.25, "len_low": 0.0}

    def default_arg_bounds(self):
        """Provide default boundaries for arguments.

        Given as a dictionary.
        """
        res = {
            "var": (0.0, 10000.0, "oc"),
            "len_scale": (0.0, 1000.0, "oo"),
            "nugget": (0.0, 100.0, "cc"),
        }
        return res

    def default_opt_arg_bounds(self):
        """Defaults for boundaries of the optional arguments.

            * ``{"hurst": [0, 1, "oo"], "len_low": [0, 1000, "cc"]}``

        Returns
        -------
        :class:`dict`
            Boundaries for optional arguments
        """
        return {"hurst": [0.1, 0.5, "oo"], "len_low": [0, 1000, "cc"]}

    def correlation(self, r):
        r"""Truncated-Power-Law with Exponential modes - correlation function.

        If ``len_low=0`` we have a simple representation:

        .. math::
           \mathrm{cor}(r) =
           H \cdot
           E_{1+H}
           \left[
           \frac{r}{\ell}
           \right]

        The general case:

        .. math::
           \mathrm{cor}(r) =
           2H \cdot
           \frac{\ell_{\mathrm{up}}^{2H} \cdot
           E_{1+2H}\left[\frac{r}{\ell_{\mathrm{up}}}\right]
           - \ell_{\mathrm{low}}^{2H} \cdot
           E_{1+2H}\left[\frac{r}{\ell_{\mathrm{low}}}\right]}
           {\ell_{\mathrm{up}}^{2H}-\ell_{\mathrm{low}}^{2H}}
        """
        # if lower limit is 0 we use the simplified version (faster)
        if np.isclose(self.len_low, 0.0):
            return tplstable_cor(r, self.len_scale, self.hurst, 1)
        return (
            self.len_up ** (2 * self.hurst)
            * tplstable_cor(r, self.len_up, self.hurst, 1)
            - self.len_low ** (2 * self.hurst)
            * tplstable_cor(r, self.len_low, self.hurst, 1)
        ) / (
            self.len_up ** (2 * self.hurst) - self.len_low ** (2 * self.hurst)
        )


class TPLStable(CovModel):
    r"""Truncated-Power-Law with Stable modes.

    Notes
    -----
    The truncated power law is given by a superposition of scale-dependent
    variograms:

    .. math::
       \gamma_{\ell_{\mathrm{low}},\ell_{\mathrm{up}}}(r) =
       \intop_{\ell_{\mathrm{low}}}^{\ell_{\mathrm{up}}}
       \gamma(r,\lambda) \frac{\rm d \lambda}{\lambda}

    with `Stable` modes on each scale:

    .. math::
       \gamma(r,\lambda) &=
       \sigma^2(\lambda)\cdot\left(1-
       \exp\left[- \left(\frac{r}{\lambda}\right)^{\alpha}\right]
       \right)\\
       \sigma^2(\lambda) &= C\cdot\lambda^{2H}

    This results in:

    .. math::
       \gamma_{\ell_{\mathrm{low}},\ell_{\mathrm{up}}}(r) &=
       \sigma^2_{\ell_{\mathrm{low}},\ell_{\mathrm{up}}}\cdot\left(1-
       \frac{2H}{\alpha} \cdot
       \frac{\ell_{\mathrm{up}}^{2H} \cdot
       E_{1+\frac{2H}{\alpha}}
       \left[\left(\frac{r}{\ell_{\mathrm{up}}}\right)^{\alpha}\right]
       - \ell_{\mathrm{low}}^{2H} \cdot
       E_{1+\frac{2H}{\alpha}}
       \left[\left(\frac{r}{\ell_{\mathrm{low}}}\right)^{\alpha}\right]}
       {\ell_{\mathrm{up}}^{2H}-\ell_{\mathrm{low}}^{2H}}
       \right) \\
       \sigma^2_{\ell_{\mathrm{low}},\ell_{\mathrm{up}}} &=
       \frac{C\cdot\left(\ell_{\mathrm{up}}^{2H}
       -\ell_{\mathrm{low}}^{2H}\right)}{2H}

    The "length scale" of this model is equivalent by the integration range:

    .. math::
       \ell = \ell_{\mathrm{up}} -\ell_{\mathrm{low}}

    If you want to define an upper scale truncation, you should set ``len_low``
    and ``len_scale`` accordingly.

    The following Parameters occure:

        * :math:`0<\alpha\leq 2` : The shape parameter of the Stable model.

            * :math:`\alpha=1` : Exponential modes
            * :math:`\alpha=2` : Gaussian modes

        * :math:`C>0` : The scaling factor from the Power-Law.
          This parameter will be calculated internally by the given variance.
          You can access C directly by ``model.var_raw``
        * :math:`0<H<\frac{\alpha}{2}` : The hurst coefficient (``model.hurst``)
        * :math:`\ell_{\mathrm{low}}\geq 0` : The lower length scale truncation
          of the model (``model.len_low``)
        * :math:`\ell_{\mathrm{up}}\geq 0` : The upper length scale truncation
          of the model (``model.len_up``)

          This will be calculated internally by:

            * ``len_up = len_low + len_scale``

          That means, that the ``len_scale`` in this model actually represents
          the integration range for the truncated power law.
        * :math:`E_s(x)` is the exponential integral.

    Other Parameters
    ----------------
    **opt_arg
        The following parameters are covered by these keyword arguments
    hurst : :class:`float`, optional
        Hurst coefficient of the power law.
        Standard range: ``(0, 1)``.
        Default: ``0.5``
    alpha : :class:`float`, optional
        Shape parameter of the stable model.
        Standard range: ``(0, 2]``.
        Default: ``1.5``
    len_low : :class:`float`, optional
        The lower length scale truncation of the model.
        Standard range: ``[0, 1000]``.
        Default: ``0.0``
    """

    @property
    def len_up(self):
        """:class:`float`: Upper length scale truncation of the model.

        * ``len_up = len_low + len_scale``
        """
        return self.len_low + self.len_scale

    def var_factor(self):
        r"""Factor for C (Power-Law factor) to result in variance.

        This is used to result in the right variance, which is depending
        on the hurst coefficient and the length-scale extents

        .. math::
           \frac{\ell_{\mathrm{up}}^{2H} - \ell_{\mathrm{low}}^{2H}}{2H}

        Returns
        -------
        :class:`float`
            factor
        """
        return (
            self.len_up ** (2 * self.hurst) - self.len_low ** (2 * self.hurst)
        ) / (2 * self.hurst)

    def default_opt_arg(self):
        """Defaults for the optional arguments.

            * ``{"hurst": 0.5, "alpha": 1.5, "len_low": 0.0}``

        Returns
        -------
        :class:`dict`
            Defaults for optional arguments
        """
        return {"hurst": 0.5, "alpha": 1.5, "len_low": 0.0}

    def default_arg_bounds(self):
        """Provide default boundaries for arguments.

        Given as a dictionary.
        """
        res = {
            "var": (0.0, 10000.0, "oc"),
            "len_scale": (0.0, 1000.0, "oo"),
            "nugget": (0.0, 100.0, "cc"),
        }
        return res

    def default_opt_arg_bounds(self):
        """Defaults for boundaries of the optional arguments.

            * ``{"hurst": [0, 1, "oo"], "alpha": [0, 2, "oc"], "len_low": [0, 1000, "cc"]}``

        Returns
        -------
        :class:`dict`
            Boundaries for optional arguments
        """
        return {
            "hurst": [0.1, 1, "oo"],
            "alpha": [0, 2, "oc"],
            "len_low": [0, 1000, "cc"],
        }

    def check_opt_arg(self):
        """Check the optional arguments.

        Warns
        -----
        alpha
            If alpha is < 0.3, the model tends to a nugget model and gets
            numerically unstable.
        """
        if self.alpha < 0.3:
            warnings.warn(
                "TPLStable: parameter 'alpha' is < 0.3, "
                + "count with unstable results"
            )

    def correlation(self, r):
        r"""Truncated-Power-Law with Stable modes - correlation function.

        If ``len_low=0`` we have a simple representation:

        .. math::
           \mathrm{cor}(r) =
           \frac{2H}{\alpha} \cdot
           E_{1+\frac{2H}{\alpha}}
           \left[
           \left(\frac{r}{\ell}\right)^{\alpha}
           \right]

        The general case:

        .. math::
           \mathrm{cor}(r) =
           \frac{2H}{\alpha} \cdot
           \frac{\ell_{\mathrm{up}}^{2H} \cdot
           E_{1+\frac{2H}{\alpha}}
           \left[\left(\frac{r}{\ell_{\mathrm{up}}}\right)^{\alpha}\right]
           - \ell_{\mathrm{low}}^{2H} \cdot
           E_{1+\frac{2H}{\alpha}}
           \left[\left(\frac{r}{\ell_{\mathrm{low}}}\right)^{\alpha}\right]}
           {\ell_{\mathrm{up}}^{2H}-\ell_{\mathrm{low}}^{2H}}
        """
        # if lower limit is 0 we use the simplified version (faster)
        if np.isclose(self.len_low, 0.0):
            return tplstable_cor(r, self.len_scale, self.hurst, self.alpha)
        return (
            self.len_up ** (2 * self.hurst)
            * tplstable_cor(r, self.len_up, self.hurst, self.alpha)
            - self.len_low ** (2 * self.hurst)
            * tplstable_cor(r, self.len_low, self.hurst, self.alpha)
        ) / (
            self.len_up ** (2 * self.hurst) - self.len_low ** (2 * self.hurst)
        )



# -*- coding: utf-8 -*-
"""
GStools subpackage providing plotting routines for the covariance models.

.. currentmodule:: gstools.covmodel.plot

The following classes and functions are provided

.. autosummary::
   plot_variogram
   plot_covariance
   plot_correlation
   plot_vario_spatial
   plot_cov_spatial
   plot_cor_spatial
   plot_spectrum
   plot_spectral_density
   plot_spectral_rad_pdf
"""
# pylint: disable=C0103
from __future__ import print_function, division, absolute_import
import numpy as np

import gstools
from gstools.field.tools import reshape_axis_from_struct_to_unstruct

__all__ = [
    "plot_variogram",
    "plot_covariance",
    "plot_correlation",
    "plot_vario_spatial",
    "plot_cov_spatial",
    "plot_cor_spatial",
    "plot_spectrum",
    "plot_spectral_density",
    "plot_spectral_rad_pdf",
]


# plotting routines #######################################################


def _get_fig_ax(fig, ax, ax_name="rectilinear"):  # pragma: no cover
    from matplotlib import pyplot as plt

    if fig is None and ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection=ax_name)
    elif ax is None:
        ax = fig.add_subplot(111, projection=ax_name)
    elif fig is None:
        fig = ax.get_figure()
        assert ax.name == ax_name
    else:
        assert ax.name == ax_name
        assert ax.get_figure() == fig
    return fig, ax


def plot_vario_spatial(
    model, x_min=0.0, x_max=None, fig=None, ax=None
):  # pragma: no cover
    """Plot spatial variogram of a given CovModel."""
    field = gstools.field.base.Field(model)
    if x_max is None:
        x_max = 3 * model.integral_scale
    field.mesh_type = "structured"
    x_s = np.linspace(-x_max, x_max) + x_min
    pos = [x_s] * model.dim
    x, y, z, shape = reshape_axis_from_struct_to_unstruct(model.dim, *pos)
    vario = model.vario_spatial([x, y, z][: model.dim]).reshape(shape)
    field.pos = pos
    field.field = vario
    return field.plot(fig=fig, ax=ax)


def plot_cov_spatial(
    model, x_min=0.0, x_max=None, fig=None, ax=None
):  # pragma: no cover
    """Plot spatial covariance of a given CovModel."""
    field = gstools.field.base.Field(model)
    if x_max is None:
        x_max = 3 * model.integral_scale
    field.mesh_type = "structured"
    x_s = np.linspace(-x_max, x_max) + x_min
    pos = [x_s] * model.dim
    x, y, z, shape = reshape_axis_from_struct_to_unstruct(model.dim, *pos)
    vario = model.cov_spatial([x, y, z][: model.dim]).reshape(shape)
    field.pos = pos
    field.field = vario
    return field.plot(fig=fig, ax=ax)


def plot_cor_spatial(
    model, x_min=0.0, x_max=None, fig=None, ax=None
):  # pragma: no cover
    """Plot spatial correlation of a given CovModel."""
    field = gstools.field.base.Field(model)
    if x_max is None:
        x_max = 3 * model.integral_scale
    field.mesh_type = "structured"
    x_s = np.linspace(-x_max, x_max) + x_min
    pos = [x_s] * model.dim
    x, y, z, shape = reshape_axis_from_struct_to_unstruct(model.dim, *pos)
    vario = model.cor_spatial([x, y, z][: model.dim]).reshape(shape)
    field.pos = pos
    field.field = vario
    return field.plot(fig=fig, ax=ax)


def plot_variogram(
    model, x_min=0.0, x_max=None, fig=None, ax=None
):  # pragma: no cover
    """Plot variogram of a given CovModel."""
    fig, ax = _get_fig_ax(fig, ax)
    if x_max is None:
        x_max = 3 * model.integral_scale
    x_s = np.linspace(x_min, x_max)
    ax.plot(x_s, model.variogram(x_s), label=model.name + " variogram")
    ax.legend()
    fig.show()
    return ax


def plot_covariance(
    model, x_min=0.0, x_max=None, fig=None, ax=None
):  # pragma: no cover
    """Plot covariance of a given CovModel."""
    fig, ax = _get_fig_ax(fig, ax)
    if x_max is None:
        x_max = 3 * model.integral_scale
    x_s = np.linspace(x_min, x_max)
    ax.plot(x_s, model.covariance(x_s), label=model.name + " covariance")
    ax.legend()
    fig.show()
    return ax


def plot_correlation(
    model, x_min=0.0, x_max=None, fig=None, ax=None
):  # pragma: no cover
    """Plot correlation function of a given CovModel."""
    fig, ax = _get_fig_ax(fig, ax)
    if x_max is None:
        x_max = 3 * model.integral_scale
    x_s = np.linspace(x_min, x_max)
    ax.plot(x_s, model.correlation(x_s), label=model.name + " correlation")
    ax.legend()
    fig.show()
    return ax


def plot_spectrum(
    model, x_min=0.0, x_max=None, fig=None, ax=None
):  # pragma: no cover
    """Plot specturm of a given CovModel."""
    fig, ax = _get_fig_ax(fig, ax)
    if x_max is None:
        x_max = 3 / model.integral_scale
    x_s = np.linspace(x_min, x_max)
    ax.plot(
        x_s,
        model.spectrum(x_s),
        label=model.name + " " + str(model.dim) + "D spectrum",
    )
    ax.legend()
    fig.show()
    return ax


def plot_spectral_density(
    model, x_min=0.0, x_max=None, fig=None, ax=None
):  # pragma: no cover
    """Plot spectral density of a given CovModel."""
    fig, ax = _get_fig_ax(fig, ax)
    if x_max is None:
        x_max = 3 / model.integral_scale
    x_s = np.linspace(x_min, x_max)
    ax.plot(
        x_s,
        model.spectral_density(x_s),
        label=model.name + " " + str(model.dim) + "D spectral-density",
    )
    ax.legend()
    fig.show()
    return ax


def plot_spectral_rad_pdf(
    model, x_min=0.0, x_max=None, fig=None, ax=None
):  # pragma: no cover
    """Plot radial spectral pdf of a given CovModel."""
    fig, ax = _get_fig_ax(fig, ax)
    if x_max is None:
        x_max = 3 / model.integral_scale
    x_s = np.linspace(x_min, x_max)
    ax.plot(
        x_s,
        model.spectral_rad_pdf(x_s),
        label=model.name + " " + str(model.dim) + "D spectral-rad-pdf",
    )
    ax.legend()
    fig.show()
    return ax



# -*- coding: utf-8 -*-
"""
GStools subpackage providing export routines.

.. currentmodule:: gstools.tools.export

The following functions are provided

.. autosummary::
   to_vtk_structured
   vtk_export_structured
   to_vtk_unstructured
   vtk_export_unstructured
   to_vtk
   vtk_export
"""
# pylint: disable=C0103, E1101
from __future__ import print_function, division, absolute_import

import numpy as np
from pyevtk.hl import gridToVTK, pointsToVTK

from gstools.tools.geometric import pos2xyz

try:
    import pyvista as pv
except ImportError:
    pv = None

__all__ = [
    "to_vtk_structured",
    "vtk_export_structured",
    "to_vtk_unstructured",
    "vtk_export_unstructured",
    "to_vtk",
    "vtk_export",
]


# export routines #############################################################


def _vtk_structured_helper(pos, fields):
    """An internal helper to extract what is needed for the vtk rectilinear grid
    """
    if not isinstance(fields, dict):
        fields = {"field": fields}
    x, y, z = pos2xyz(pos)
    if y is None:
        y = np.array([0])
    if z is None:
        z = np.array([0])
    # need fortran order in VTK
    for field in fields:
        fields[field] = fields[field].reshape(-1, order="F")
        if len(fields[field]) != len(x) * len(y) * len(z):
            raise ValueError(
                "gstools.vtk_export_structured: "
                "field shape doesn't match the given mesh"
            )
    return x, y, z, fields


def to_vtk_structured(pos, fields):  # pragma: no cover
    """Create a vtk structured rectilinear grid from a field.

    Parameters
    ----------
    pos : :class:`list`
        the position tuple, containing main direction and transversal
        directions
    fields : :class:`dict` or :class:`numpy.ndarray`
        Structured fields to be saved.
        Either a single numpy array as returned by SRF,
        or a dictionary of fields with theirs names as keys.

    Returns
    -------
    :class:`pyvista.RectilinearGrid`
        A PyVista rectilinear grid of the structured field data. Data arrays
        live on the point data of this PyVista dataset.
    """
    x, y, z, fields = _vtk_structured_helper(pos=pos, fields=fields)
    try:
        import pyvista as pv

        grid = pv.RectilinearGrid(x, y, z)
        grid.point_arrays.update(fields)
    except ImportError:
        raise ImportError("Please install PyVista to create VTK datasets.")
    return grid


def vtk_export_structured(filename, pos, fields):  # pragma: no cover
    """Export a field to vtk structured rectilinear grid file.

    Parameters
    ----------
    filename : :class:`str`
        Filename of the file to be saved, including the path. Note that an
        ending (.vtr) will be added to the name.
    pos : :class:`list`
        the position tuple, containing main direction and transversal
        directions
    fields : :class:`dict` or :class:`numpy.ndarray`
        Structured fields to be saved.
        Either a single numpy array as returned by SRF,
        or a dictionary of fields with theirs names as keys.
    """
    x, y, z, fields = _vtk_structured_helper(pos=pos, fields=fields)
    return gridToVTK(filename, x, y, z, pointData=fields)


def _vtk_unstructured_helper(pos, fields):
    if not isinstance(fields, dict):
        fields = {"field": fields}
    x, y, z = pos2xyz(pos)
    if y is None:
        y = np.zeros_like(x)
    if z is None:
        z = np.zeros_like(x)
    for field in fields:
        fields[field] = fields[field].reshape(-1)
        if (
            len(fields[field]) != len(x)
            or len(fields[field]) != len(y)
            or len(fields[field]) != len(z)
        ):
            raise ValueError(
                "gstools.vtk_export_unstructured: "
                "field shape doesn't match the given mesh"
            )
    return x, y, z, fields


def to_vtk_unstructured(pos, fields):  # pragma: no cover
    """Export a field to vtk structured rectilinear grid file.

    Parameters
    ----------
    pos : :class:`list`
        the position tuple, containing main direction and transversal
        directions
    fields : :class:`dict` or :class:`numpy.ndarray`
        Unstructured fields to be saved.
        Either a single numpy array as returned by SRF,
        or a dictionary of fields with theirs names as keys.

    Returns
    -------
    :class:`pyvista.UnstructuredGrid`
        A PyVista unstructured grid of the unstructured field data. Data arrays
        live on the point data of this PyVista dataset. This is essentially
        a point cloud with no topology.
    """
    x, y, z, fields = _vtk_unstructured_helper(pos=pos, fields=fields)
    try:
        import pyvista as pv

        grid = pv.PolyData(np.c_[x, y, z]).cast_to_unstructured_grid()
        grid.point_arrays.update(fields)
    except ImportError:
        raise ImportError("Please install PyVista to create VTK datasets.")
    return grid


def vtk_export_unstructured(filename, pos, fields):  # pragma: no cover
    """Export a field to vtk unstructured grid file.

    Parameters
    ----------
    filename : :class:`str`
        Filename of the file to be saved, including the path. Note that an
        ending (.vtu) will be added to the name.
    pos : :class:`list`
        the position tuple, containing main direction and transversal
        directions
    fields : :class:`dict` or :class:`numpy.ndarray`
        Unstructured fields to be saved.
        Either a single numpy array as returned by SRF,
        or a dictionary of fields with theirs names as keys.
    """
    x, y, z, fields = _vtk_unstructured_helper(pos=pos, fields=fields)
    return pointsToVTK(filename, x, y, z, data=fields)


def to_vtk(pos, fields, mesh_type="unstructured"):  # pragma: no cover
    """Create a VTK/PyVista grid.

    Parameters
    ----------
    pos : :class:`list`
        the position tuple, containing main direction and transversal
        directions
    fields : :class:`dict` or :class:`numpy.ndarray`
        [Un]structured fields to be saved.
        Either a single numpy array as returned by SRF,
        or a dictionary of fields with theirs names as keys.
    mesh_type : :class:`str`, optional
        'structured' / 'unstructured'. Default: structured

    Returns
    -------
    :class:`pyvista.RectilinearGrid` or :class:`pyvista.UnstructuredGrid`
        This will return a PyVista object for the given field data in its
        appropriate type. Structured meshes will return a
        :class:`pyvista.RectilinearGrid` and unstructured meshes will return
        an :class:`pyvista.UnstructuredGrid` object.
    """
    if mesh_type == "structured":
        grid = to_vtk_structured(pos=pos, fields=fields)
    else:
        grid = to_vtk_unstructured(pos=pos, fields=fields)
    return grid


def vtk_export(
    filename, pos, fields, mesh_type="unstructured"
):  # pragma: no cover
    """Export a field to vtk.

    Parameters
    ----------
    filename : :class:`str`
        Filename of the file to be saved, including the path. Note that an
        ending (.vtr or .vtu) will be added to the name.
    pos : :class:`list`
        the position tuple, containing main direction and transversal
        directions
    fields : :class:`dict` or :class:`numpy.ndarray`
        [Un]structured fields to be saved.
        Either a single numpy array as returned by SRF,
        or a dictionary of fields with theirs names as keys.
    mesh_type : :class:`str`, optional
        'structured' / 'unstructured'. Default: structured
    """
    if mesh_type == "structured":
        return vtk_export_structured(filename=filename, pos=pos, fields=fields)
    else:
        return vtk_export_unstructured(
            filename=filename, pos=pos, fields=fields
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This is a unittest of the variogram module.
"""
from __future__ import division, absolute_import, print_function

import sys
import unittest
import numpy as np
from gstools import vario_estimate_unstructured

PY3 = sys.version_info[0] == 3
# in python3 "long" was replaced with "int"
# https://docs.python.org/3.3/whatsnew/3.0.html#integers
if PY3:
    LONGTYPE = int
else:
    LONGTYPE = long


class TestVariogramUnstructured(unittest.TestCase):
    def setUp(self):
        pass

    def test_doubles(self):
        x = np.arange(1, 11, 1, dtype=np.double)
        z = np.array(
            (41.2, 40.2, 39.7, 39.2, 40.1, 38.3, 39.1, 40.0, 41.1, 40.3),
            dtype=np.double,
        )
        bins = np.arange(1, 11, 1, dtype=np.double)
        bin_centres, gamma = vario_estimate_unstructured([x], z, bins)
        self.assertAlmostEqual(gamma[0], 0.4917, places=4)

    def test_ints(self):
        x = np.arange(1, 5, 1, dtype=int)
        z = np.array((10, 20, 30, 40), dtype=int)
        bins = np.arange(1, 11, 1, dtype=int)
        bin_centres, gamma = vario_estimate_unstructured([x], z, bins)
        self.assertAlmostEqual(gamma[0], 50.0, places=4)

    def test_longs(self):
        x = np.arange(1, 5, 1, dtype=LONGTYPE)
        z = np.array((10, 20, 30, 40), dtype=LONGTYPE)
        bins = np.arange(1, 11, 1, dtype=LONGTYPE)
        bin_centres, gamma = vario_estimate_unstructured([x], z, bins)
        self.assertAlmostEqual(gamma[0], 50.0, places=4)

    def test_np_int(self):
        x = np.arange(1, 5, 1, dtype=np.int)
        z = np.array((10, 20, 30, 40), dtype=np.int)
        bins = np.arange(1, 11, 1, dtype=np.int)
        bin_centres, gamma = vario_estimate_unstructured([x], z, bins)
        self.assertAlmostEqual(gamma[0], 50.0, places=4)

    def test_mixed(self):
        x = np.arange(1, 11, 1, dtype=np.double)
        z = np.array(
            (41.2, 40.2, 39.7, 39.2, 40.1, 38.3, 39.1, 40.0, 41.1, 40.3),
            dtype=np.double,
        )
        bins = np.arange(1, 11, 1, dtype=int)
        bin_centres, gamma = vario_estimate_unstructured([x], z, bins)
        self.assertAlmostEqual(gamma[0], 0.4917, places=4)

        x = np.arange(1, 5, 1, dtype=np.double)
        z = np.array((10, 20, 30, 40), dtype=LONGTYPE)
        bins = np.arange(1, 11, 1, dtype=int)
        bin_centres, gamma = vario_estimate_unstructured([x], z, bins)
        self.assertAlmostEqual(gamma[0], 50.0, places=4)

        x = np.arange(1, 5, 1, dtype=np.double)
        z = np.array((10, 20, 30, 40), dtype=LONGTYPE)
        bins = np.arange(1, 11, 1, dtype=np.double)
        bin_centres, gamma = vario_estimate_unstructured([x], z, bins)
        self.assertAlmostEqual(gamma[0], 50.0, places=4)

    def test_list(self):
        x = np.arange(1, 11, 1, dtype=np.double)
        z = [41.2, 40.2, 39.7, 39.2, 40.1, 38.3, 39.1, 40.0, 41.1, 40.3]
        bins = np.arange(1, 11, 1, dtype=np.double)
        bin_centres, gamma = vario_estimate_unstructured([x], z, bins)
        self.assertAlmostEqual(gamma[1], 0.7625, places=4)

    def test_1d(self):
        x = np.arange(1, 11, 1, dtype=np.double)
        # literature values
        z = np.array(
            (41.2, 40.2, 39.7, 39.2, 40.1, 38.3, 39.1, 40.0, 41.1, 40.3),
            dtype=np.double,
        )
        bins = np.arange(1, 11, 1, dtype=np.double)
        bin_centres, gamma = vario_estimate_unstructured([x], z, bins)
        self.assertAlmostEqual(gamma[0], 0.4917, places=4)
        self.assertAlmostEqual(gamma[1], 0.7625, places=4)

    def test_uncorrelated_2d(self):
        x_c = np.linspace(0.0, 100.0, 60)
        y_c = np.linspace(0.0, 100.0, 60)
        x, y = np.meshgrid(x_c, y_c)
        x = np.reshape(x, len(x_c) * len(y_c))
        y = np.reshape(y, len(x_c) * len(y_c))

        rng = np.random.RandomState(1479373475)
        field = rng.rand(len(x))

        bins = np.arange(0, 100, 10)

        bin_centres, gamma = vario_estimate_unstructured((x, y), field, bins)

        var = 1.0 / 12.0
        self.assertAlmostEqual(gamma[0], var, places=2)
        self.assertAlmostEqual(gamma[len(gamma) // 2], var, places=2)
        self.assertAlmostEqual(gamma[-1], var, places=2)

    def test_uncorrelated_3d(self):
        x_c = np.linspace(0.0, 100.0, 15)
        y_c = np.linspace(0.0, 100.0, 15)
        z_c = np.linspace(0.0, 100.0, 15)
        x, y, z = np.meshgrid(x_c, y_c, z_c)
        x = np.reshape(x, len(x_c) * len(y_c) * len(z_c))
        y = np.reshape(y, len(x_c) * len(y_c) * len(z_c))
        z = np.reshape(z, len(x_c) * len(y_c) * len(z_c))

        rng = np.random.RandomState(1479373475)
        field = rng.rand(len(x))

        bins = np.arange(0, 100, 10)

        bin_centres, gamma = vario_estimate_unstructured(
            (x, y, z), field, bins
        )

        var = 1.0 / 12.0
        self.assertAlmostEqual(gamma[0], var, places=2)
        self.assertAlmostEqual(gamma[len(gamma) // 2], var, places=2)
        self.assertAlmostEqual(gamma[-1], var, places=2)

    def test_sampling_1d(self):
        x = np.linspace(0.0, 100.0, 21000)

        rng = np.random.RandomState(1479373475)
        field = rng.rand(len(x))

        bins = np.arange(0, 100, 10)

        bin_centres, gamma = vario_estimate_unstructured(
            [x], field, bins, sampling_size=5000, sampling_seed=1479373475
        )

        var = 1.0 / 12.0
        self.assertAlmostEqual(gamma[0], var, places=2)
        self.assertAlmostEqual(gamma[len(gamma) // 2], var, places=2)
        self.assertAlmostEqual(gamma[-1], var, places=2)

    def test_sampling_2d(self):
        x_c = np.linspace(0.0, 100.0, 600)
        y_c = np.linspace(0.0, 100.0, 600)
        x, y = np.meshgrid(x_c, y_c)
        x = np.reshape(x, len(x_c) * len(y_c))
        y = np.reshape(y, len(x_c) * len(y_c))

        rng = np.random.RandomState(1479373475)
        field = rng.rand(len(x))

        bins = np.arange(0, 100, 10)

        bin_centres, gamma = vario_estimate_unstructured(
            (x, y), field, bins, sampling_size=2000, sampling_seed=1479373475
        )

        var = 1.0 / 12.0
        self.assertAlmostEqual(gamma[0], var, places=2)
        self.assertAlmostEqual(gamma[len(gamma) // 2], var, places=2)
        self.assertAlmostEqual(gamma[-1], var, places=2)

    def test_sampling_3d(self):
        x_c = np.linspace(0.0, 100.0, 100)
        y_c = np.linspace(0.0, 100.0, 100)
        z_c = np.linspace(0.0, 100.0, 100)
        x, y, z = np.meshgrid(x_c, y_c, z_c)
        x = np.reshape(x, len(x_c) * len(y_c) * len(z_c))
        y = np.reshape(y, len(x_c) * len(y_c) * len(z_c))
        z = np.reshape(z, len(x_c) * len(y_c) * len(z_c))

        rng = np.random.RandomState(1479373475)
        field = rng.rand(len(x))

        bins = np.arange(0, 100, 10)

        bin_centres, gamma = vario_estimate_unstructured(
            (x, y, z),
            field,
            bins,
            sampling_size=2000,
            sampling_seed=1479373475,
        )
        var = 1.0 / 12.0
        self.assertAlmostEqual(gamma[0], var, places=2)
        self.assertAlmostEqual(gamma[len(gamma) // 2], var, places=2)
        self.assertAlmostEqual(gamma[-1], var, places=2)

    def test_assertions(self):
        x = np.arange(0, 10)
        x_e = np.arange(0, 11)
        y = np.arange(0, 11)
        y_e = np.arange(0, 12)
        z = np.arange(0, 12)
        z_e = np.arange(0, 15)
        bins = np.arange(0, 3)
        #        bins_e = np.arange(0, 1)
        field = np.arange(0, 10)
        field_e = np.arange(0, 9)

        self.assertRaises(
            ValueError, vario_estimate_unstructured, [x_e], field, bins
        )
        self.assertRaises(
            ValueError, vario_estimate_unstructured, (x, y_e), field, bins
        )
        self.assertRaises(
            ValueError, vario_estimate_unstructured, (x, y_e, z), field, bins
        )
        self.assertRaises(
            ValueError, vario_estimate_unstructured, (x, y, z_e), field, bins
        )
        self.assertRaises(
            ValueError, vario_estimate_unstructured, (x_e, y, z), field, bins
        )
        self.assertRaises(
            ValueError, vario_estimate_unstructured, (x, y, z), field_e, bins
        )
        self.assertRaises(
            ValueError, vario_estimate_unstructured, [x], field_e, bins
        )


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
GStools subpackage providing the core of the spatial random field generation.

.. currentmodule:: gstools.random.rng

The following classes are provided

.. autosummary::
   RNG
"""
# pylint: disable=no-member
from __future__ import division, absolute_import, print_function

import numpy as np
import numpy.random as rand
import emcee as mc
from emcee.state import State
from gstools.random.tools import MasterRNG, dist_gen

__all__ = ["RNG"]


class RNG(object):
    """
    A random number generator for different distributions and multiple streams.

    Parameters
    ----------
    seed : :class:`int` or :any:`None`, optional
        The seed of the master RNG, if ``None``,
        a random seed is used. Default: ``None``
    """

    def __init__(self, seed=None):
        # set seed
        self._master_rng = None
        self.seed = seed

    def sample_ln_pdf(
        self,
        ln_pdf,
        size=None,
        sample_around=1.0,
        nwalkers=50,
        burn_in=20,
        oversampling_factor=10,
    ):
        """Sample from a distribution given by ln(pdf).

        This algorithm uses the :class:`emcee.EnsembleSampler`

        Parameters
        ----------
        ln_pdf : :any:`callable`
            The logarithm of the Probability density function
            of the given distribution, that takes a single argument
        size : :class:`int` or :any:`None`, optional
            sample size. Default: None
        sample_around : :class:`float`, optional
            Starting point for initial guess Default: 1.
        nwalkers : :class:`int`, optional
            The number of walkers in the mcmc sampler. Used for the
            emcee.EnsembleSampler class.
            Default: 100
        burn_in : :class:`int`, optional
            Number of burn-in runs in the mcmc algorithm.
            Default: 100
        oversampling_factor : :class:`int`, optional
            To guess the sample number needed for proper results, we use a
            factor for oversampling. The intern used sample-size is
            calculated by

            ``sample_size = max(burn_in, (size/nwalkers)*oversampling_factor)``

            So at least, as much as the burn-in runs.
            Default: 10
        """
        if size is None:  # pragma: no cover
            sample_size = burn_in
        else:
            sample_size = max(burn_in, (size / nwalkers) * oversampling_factor)
        # initial guess
        init_guess = (
            self.random.rand(nwalkers).reshape((nwalkers, 1)) * sample_around
        )
        # initialize the sampler
        sampler = mc.EnsembleSampler(nwalkers, 1, ln_pdf)
        # burn in phase with saving of last position
        initial_state = State(init_guess, copy=True)
        initial_state.random_state = self.random.get_state()
        burn_in_state = sampler.run_mcmc(
            initial_state=initial_state, nsteps=burn_in
        )
        # reset after burn_in
        sampler.reset()
        # actual sampling
        initial_state = State(burn_in_state, copy=True)
        initial_state.random_state = self.random.get_state()
        sampler.run_mcmc(initial_state=initial_state, nsteps=sample_size)
        samples = sampler.get_chain(flat=True)[:, 0]

        # choose samples according to size
        return self.random.choice(samples, size)

    def sample_dist(self, pdf=None, cdf=None, ppf=None, size=None, **kwargs):
        """Sample from a distribution given by pdf, cdf and/or ppf.

        Parameters
        ----------
        pdf : :any:`callable` or :any:`None`, optional
            Probability density function of the given distribution,
            that takes a single argument
            Default: ``None``
        cdf : :any:`callable` or :any:`None`, optional
            Cumulative distribution function of the given distribution, that
            takes a single argument
            Default: ``None``
        ppf : :any:`callable` or :any:`None`, optional
            Percent point function of the given distribution, that
            takes a single argument
            Default: ``None``
        size : :class:`int` or :any:`None`, optional
            sample size. Default: None
        **kwargs
            Keyword-arguments that are forwarded to
            :any:`scipy.stats.rv_continuous`.

        Returns
        -------
        samples : :class:`float` or :class:`numpy.ndarray`
            the samples from the given distribution

        Notes
        -----
        At least pdf or cdf needs to be given.
        """
        kwargs["seed"] = self.random
        dist = dist_gen(pdf_in=pdf, cdf_in=cdf, ppf_in=ppf, **kwargs)
        return dist.rvs(size=size)

    def sample_sphere(self, dim, size=None):
        """Uniform sampling on a d-dimensional sphere.

        Parameters
        ----------
        dim : :class:`int`
            Dimension of the sphere. Just 1, 2, and 3 supported.
        size : :class:`int`, optional
            sample size

        Returns
        -------
        coord : :class:`numpy.ndarray`
            x[, y[, z]] coordinates on the sphere with shape (dim, size)
        """
        if size is None:  # pragma: no cover
            coord = np.empty(dim, dtype=np.double)
        else:
            coord = np.empty((dim, size), dtype=np.double)
        if dim == 1:
            coord[0] = self.random.choice([-1, 1], size=size)
        elif dim == 2:
            ang1 = self.random.uniform(0.0, 2 * np.pi, size)
            coord[0] = np.cos(ang1)
            coord[1] = np.sin(ang1)
        elif dim == 3:
            ang1 = self.random.uniform(0.0, 2 * np.pi, size)
            ang2 = self.random.uniform(-1.0, 1.0, size)
            coord[0] = np.sqrt(1.0 - ang2 ** 2) * np.cos(ang1)
            coord[1] = np.sqrt(1.0 - ang2 ** 2) * np.sin(ang1)
            coord[2] = ang2
        return coord

    @property
    def random(self):
        """:any:`numpy.random.mtrand.RandomState`: Randomstate.

        Get a stream to the numpy Random number generator.
        You can use this, to call any provided distribution
        from :any:`numpy.random.mtrand.RandomState`.
        """
        return rand.RandomState(self._master_rng())

    @property  # pragma: no cover
    def seed(self):
        """:class:`int`: Seed of the master RNG.

        The setter property not only saves the new seed, but also creates
        a new master RNG function with the new seed.
        """
        return self._master_rng.seed

    @seed.setter
    def seed(self, new_seed=None):
        self._master_rng = MasterRNG(new_seed)

    def __str__(self):
        """Return String representation."""
        return self.__repr__()

    def __repr__(self):
        """Return String representation."""
        return "RNG(seed={})".format(self.seed)


if __name__ == "__main__":  # pragma: no cover
    import doctest

    doctest.testmod()



# -*- coding: utf-8 -*-
"""
GStools subpackage providing different covariance models.

.. currentmodule:: gstools.covmodel.models

The following classes and functions are provided

.. autosummary::
   Gaussian
   Exponential
   Matern
   Stable
   Rational
   Linear
   Circular
   Spherical
   Intersection
"""
# pylint: disable=C0103, E1101, E1137
from __future__ import print_function, division, absolute_import

import warnings
import numpy as np
from scipy import special as sps
from gstools.covmodel.base import CovModel

__all__ = [
    "Gaussian",
    "Exponential",
    "Matern",
    "Stable",
    "Rational",
    "Linear",
    "Circular",
    "Spherical",
    "Intersection",
]


# Gaussian Model ##############################################################


class Gaussian(CovModel):
    r"""The Gaussian covariance model.

    Notes
    -----
    This model is given by the following correlation function:

    .. math::
       \mathrm{cor}(r) =
       \exp\left(- \frac{\pi}{4} \cdot \left(\frac{r}{\ell}\right)^2\right)

    """

    def correlation(self, r):
        r"""Gaussian correlation function.

        .. math::
           \mathrm{cor}(r) =
           \exp\left(- \frac{\pi}{4}\cdot \left(\frac{r}{\ell}\right)^2\right)
        """
        r = np.array(np.abs(r), dtype=np.double)
        return np.exp(-np.pi / 4 * (r / self.len_scale) ** 2)

    def spectrum(self, k):  # noqa: D102
        k = np.array(k, dtype=np.double)
        return (
            self.var
            * (self.len_scale / np.pi) ** self.dim
            * np.exp(-(k * self.len_scale) ** 2 / np.pi)
        )

    def spectral_rad_cdf(self, r):
        """Radial spectral cdf."""
        r = np.array(r, dtype=np.double)
        if self.dim == 1:
            return sps.erf(self.len_scale * r / np.sqrt(np.pi))
        if self.dim == 2:
            return 1.0 - np.exp(-(r * self.len_scale) ** 2 / np.pi)
        if self.dim == 3:
            return sps.erf(
                self.len_scale * r / np.sqrt(np.pi)
            ) - 2 * r * self.len_scale / np.pi * np.exp(
                -(r * self.len_scale) ** 2 / np.pi
            )
        return None

    def spectral_rad_ppf(self, u):
        """Radial spectral ppf.

        Notes
        -----
        Not defined for 3D.
        """
        u = np.array(u, dtype=np.double)
        if self.dim == 1:
            return sps.erfinv(u) * np.sqrt(np.pi) / self.len_scale
        if self.dim == 2:
            return np.sqrt(np.pi) / self.len_scale * np.sqrt(-np.log(1.0 - u))
        return None

    def _has_ppf(self):
        # since the ppf is not analytical for dim=3, we have to state that
        if self.dim == 3:
            return False
        return True

    def calc_integral_scale(self):  # noqa: D102
        return self.len_scale


# Exponential Model ###########################################################


class Exponential(CovModel):
    r"""The Exponential covariance model.

    Notes
    -----
    This model is given by the following correlation function:

    .. math::
       \mathrm{cor}(r) =
       \exp\left(- \frac{r}{\ell} \right)

    """

    def correlation(self, r):
        r"""Exponential correlation function.

        .. math::
           \mathrm{cor}(r) =
           \exp\left(- \frac{r}{\ell} \right)
        """
        r = np.array(np.abs(r), dtype=np.double)
        return np.exp(-1 * r / self.len_scale)

    def spectrum(self, k):  # noqa: D102
        k = np.array(k, dtype=np.double)
        return (
            self.var
            * self.len_scale ** self.dim
            * sps.gamma((self.dim + 1) / 2)
            / (np.pi * (1.0 + (k * self.len_scale) ** 2))
            ** ((self.dim + 1) / 2)
        )

    def spectral_rad_cdf(self, r):
        """Radial spectral cdf."""
        r = np.array(r, dtype=np.double)
        if self.dim == 1:
            return np.arctan(r * self.len_scale) * 2 / np.pi
        if self.dim == 2:
            return 1.0 - 1 / np.sqrt(1 + (r * self.len_scale) ** 2)
        if self.dim == 3:
            return (
                (
                    np.arctan(r * self.len_scale)
                    - r * self.len_scale / (1 + (r * self.len_scale) ** 2)
                )
                * 2
                / np.pi
            )
        return None

    def spectral_rad_ppf(self, u):
        """Radial spectral ppf.

        Notes
        -----
        Not defined for 3D.
        """
        u = np.array(u, dtype=np.double)
        if self.dim == 1:
            return np.tan(np.pi / 2 * u) / self.len_scale
        if self.dim == 2:
            u_power = np.divide(
                1,
                u ** 2,
                out=np.full_like(u, np.inf),
                where=np.logical_not(np.isclose(u, 0)),
            )
            return np.sqrt(u_power - 1.0) / self.len_scale
        return None

    def _has_ppf(self):
        # since the ppf is not analytical for dim=3, we have to state that
        if self.dim == 3:
            return False
        return True

    def calc_integral_scale(self):  # noqa: D102
        return self.len_scale


# Rational Model ##############################################################


class Rational(CovModel):
    r"""The rational quadratic covariance model.

    Notes
    -----
    This model is given by the following correlation function:

    .. math::
       \mathrm{cor}(r) =
       \left(1 + \frac{1}{2\alpha} \cdot
       \left(\frac{r}{\ell}\right)^2\right)^{-\alpha}

    :math:`\alpha` is a shape parameter and should be > 0.5.

    Other Parameters
    ----------------
    **opt_arg
        The following parameters are covered by these keyword arguments
    alpha : :class:`float`, optional
        Shape parameter. Standard range: ``(0, inf)``
        Default: ``1.0``
    """

    def default_opt_arg(self):
        """Defaults for the optional arguments.

            * ``{"alpha": 1.0}``

        Returns
        -------
        :class:`dict`
            Defaults for optional arguments
        """
        return {"alpha": 1.0}

    def default_opt_arg_bounds(self):
        """Defaults for boundaries of the optional arguments.

            * ``{"alpha": [0.5, inf]}``

        Returns
        -------
        :class:`dict`
            Boundaries for optional arguments
        """
        return {"alpha": [0.5, np.inf]}

    def correlation(self, r):
        r"""Rational correlation function.

        .. math::
           \mathrm{cor}(r) =
           \left(1 + \frac{1}{2\alpha} \cdot
           \left(\frac{r}{\ell}\right)^2\right)^{-\alpha}
        """
        r = np.array(np.abs(r), dtype=np.double)
        return np.power(
            1 + 0.5 / self.alpha * (r / self.len_scale) ** 2, -self.alpha
        )


# Stable Model ################################################################


class Stable(CovModel):
    r"""The stable covariance model.

    Notes
    -----
    This model is given by the following correlation function:

    .. math::
       \mathrm{cor}(r) =
       \exp\left(- \left(\frac{r}{\ell}\right)^{\alpha}\right)

    :math:`\alpha` is a shape parameter with :math:`\alpha\in(0,2]`

    Other Parameters
    ----------------
    **opt_arg
        The following parameters are covered by these keyword arguments
    alpha : :class:`float`, optional
        Shape parameter. Standard range: ``(0, 2]``
        Default: ``1.5``
    """

    def default_opt_arg(self):
        """Defaults for the optional arguments.

            * ``{"alpha": 1.5}``

        Returns
        -------
        :class:`dict`
            Defaults for optional arguments
        """
        return {"alpha": 1.5}

    def default_opt_arg_bounds(self):
        """Defaults for boundaries of the optional arguments.

            * ``{"alpha": [0, 2, "oc"]}``

        Returns
        -------
        :class:`dict`
            Boundaries for optional arguments
        """
        return {"alpha": [0, 2, "oc"]}

    def check_opt_arg(self):
        """Check the optional arguments.

        Warns
        -----
        alpha
            If alpha is < 0.3, the model tends to a nugget model and gets
            numerically unstable.
        """
        if self.alpha < 0.3:
            warnings.warn(
                "Stable: parameter 'alpha' is < 0.3, "
                + "count with unstable results"
            )

    def correlation(self, r):
        r"""Stable correlation function.

        .. math::
           \mathrm{cor}(r) =
           \exp\left(- \left(\frac{r}{\ell}\right)^{\alpha}\right)
        """
        r = np.array(np.abs(r), dtype=np.double)
        return np.exp(-np.power(r / self.len_scale, self.alpha))


# Matérn Model ################################################################


class Matern(CovModel):
    r"""The Matérn covariance model.

    Notes
    -----
    This model is given by the following correlation function:

    .. math::
       \mathrm{cor}(r) =
       \frac{2^{1-\nu}}{\Gamma\left(\nu\right)} \cdot
       \left(\sqrt{\nu}\cdot\frac{r}{\ell}\right)^{\nu} \cdot
       \mathrm{K}_{\nu}\left(\sqrt{\nu}\cdot\frac{r}{\ell}\right)

    Where :math:`\Gamma` is the gamma function and :math:`\mathrm{K}_{\nu}`
    is the modified Bessel function of the second kind.

    :math:`\nu` is a shape parameter and should be >= 0.2.

    If :math:`\nu > 20`, a gaussian model is used, since it is the limit
    case:

    .. math::
       \mathrm{cor}(r) =
       \exp\left(- \frac{1}{4} \cdot \left(\frac{r}{\ell}\right)^2\right)

    Other Parameters
    ----------------
    **opt_arg
        The following parameters are covered by these keyword arguments
    nu : :class:`float`, optional
        Shape parameter. Standard range: ``[0.2, 30]``
        Default: ``1.0``
    """

    def default_opt_arg(self):
        """Defaults for the optional arguments.

            * ``{"nu": 1.0}``

        Returns
        -------
        :class:`dict`
            Defaults for optional arguments
        """
        return {"nu": 1.0}

    def default_opt_arg_bounds(self):
        """Defaults for boundaries of the optional arguments.

            * ``{"nu": [0.5, 30.0, "cc"]}``

        Returns
        -------
        :class:`dict`
            Boundaries for optional arguments
        """
        return {"nu": [0.2, 30.0, "cc"]}

    def correlation(self, r):
        r"""Matérn correlation function.

        .. math::
           \mathrm{cor}(r) =
           \frac{2^{1-\nu}}{\Gamma\left(\nu\right)} \cdot
           \left(\sqrt{\nu}\cdot\frac{r}{\ell}\right)^{\nu} \cdot
           \mathrm{K}_{\nu}\left(\sqrt{\nu}\cdot\frac{r}{\ell}\right)
        """
        r = np.array(np.abs(r), dtype=np.double)
        # for nu > 20 we just use the gaussian model
        if self.nu > 20.0:
            return np.exp(-(r / self.len_scale) ** 2 / 4)
        # calculate by log-transformation to prevent numerical errors
        r_gz = r[r > 0.0]
        res = np.ones_like(r)
        # with np.errstate(over="ignore", invalid="ignore"):
        res[r > 0.0] = np.exp(
            (1.0 - self.nu) * np.log(2)
            - sps.loggamma(self.nu)
            + self.nu * np.log(np.sqrt(self.nu) * r_gz / self.len_scale)
        ) * sps.kv(self.nu, np.sqrt(self.nu) * r_gz / self.len_scale)
        # if nu >> 1 we get errors for the farfield, there 0 is approached
        res[np.logical_not(np.isfinite(res))] = 0.0
        # covariance is positiv
        res = np.maximum(res, 0.0)
        return res

    def spectrum(self, k):  # noqa: D102
        k = np.array(k, dtype=np.double)
        # for nu > 20 we just use an approximation of the gaussian model
        if self.nu > 20.0:
            return (
                self.var
                * (self.len_scale / np.sqrt(np.pi)) ** self.dim
                * np.exp(-(k * self.len_scale) ** 2)
                * (
                    1
                    + (
                        ((k * self.len_scale) ** 2 - self.dim / 2.0) ** 2
                        - self.dim / 2.0
                    )
                    / self.nu
                )
            )
        return (
            self.var
            * (self.len_scale / np.sqrt(np.pi)) ** self.dim
            * np.exp(
                -(self.nu + self.dim / 2.0)
                * np.log(1.0 + (k * self.len_scale) ** 2 / self.nu)
                + sps.loggamma(self.nu + self.dim / 2.0)
                - sps.loggamma(self.nu)
                - self.dim * np.log(np.sqrt(self.nu))
            )
        )

    def calc_integral_scale(self):  # noqa: D102
        return (
            self.len_scale * np.pi / np.sqrt(self.nu) / sps.beta(self.nu, 0.5)
        )


# Bounded linear Model ########################################################


class Linear(CovModel):
    r"""The bounded linear covariance model.

    This model is derived from the relative intersection area of
    two lines in 1D, where the middle points have a distance of :math:`r`
    and the line lengths are :math:`\ell`.

    Notes
    -----
    This model is given by the following correlation function:

    .. math::
       \mathrm{cor}(r) =
       \begin{cases}
       1-\frac{r}{\ell}
       & r<\ell\\
       0 & r\geq\ell
       \end{cases}

    """

    def correlation(self, r):
        r"""Linear correlation function.

        .. math::
           \mathrm{cor}(r) =
           \begin{cases}
           1-\frac{r}{\ell}
           & r<\ell\\
           0 & r\geq\ell
           \end{cases}
        """
        r = np.array(np.abs(r), dtype=np.double)
        res = np.zeros_like(r)
        r_ll = r < self.len_scale
        r_low = r[r_ll]
        res[r_ll] = 1.0 - r_low / self.len_scale
        return res


# Circular Model ##############################################################


class Circular(CovModel):
    r"""The circular covariance model.

    This model is derived as the relative intersection area of
    two discs in 2D, where the middle points have a distance of :math:`r`
    and the diameters are given by :math:`\ell`.

    Notes
    -----
    This model is given by the following correlation function:

    .. math::
       \mathrm{cor}(r) =
       \begin{cases}
       \frac{2}{\pi}\cdot\left(
       \cos^{-1}\left(\frac{r}{\ell}\right) -
       \frac{r}{\ell}\cdot\sqrt{1-\left(\frac{r}{\ell}\right)^{2}}
       \right)
       & r<\ell\\
       0 & r\geq\ell
       \end{cases}

    """

    def correlation(self, r):
        r"""Circular correlation function.

        .. math::
           \mathrm{cor}(r) =
           \begin{cases}
           \frac{2}{\pi}\cdot\left(
           \cos^{-1}\left(\frac{r}{\ell}\right) -
           \frac{r}{\ell}\cdot\sqrt{1-\left(\frac{r}{\ell}\right)^{2}}
           \right)
           & r<\ell\\
           0 & r\geq\ell
           \end{cases}

        """
        r = np.array(np.abs(r), dtype=np.double)
        res = np.zeros_like(r)
        r_ll = r < self.len_scale
        r_low = r[r_ll]
        res[r_ll] = (
            2
            / np.pi
            * (
                np.arccos(r_low / self.len_scale)
                - r_low
                / self.len_scale
                * np.sqrt(1 - (r_low / self.len_scale) ** 2)
            )
        )
        return res


# Spherical Model #############################################################


class Spherical(CovModel):
    r"""The Spherical covariance model.

    This model is derived from the relative intersection area of
    two spheres in 3D, where the middle points have a distance of :math:`r`
    and the diameters are given by :math:`\ell`.

    Notes
    -----
    This model is given by the following correlation function:

    .. math::
       \mathrm{cor}(r) =
       \begin{cases}
       1-\frac{3}{2}\cdot\frac{r}{\ell} +
       \frac{1}{2}\cdot\left(\frac{r}{\ell}\right)^{3}
       & r<\ell\\
       0 & r\geq\ell
       \end{cases}

    """

    def correlation(self, r):
        r"""Spherical correlation function.

        .. math::
           \mathrm{cor}(r) =
           \begin{cases}
           1-\frac{3}{2}\cdot\frac{r}{\ell} +
           \frac{1}{2}\cdot\left(\frac{r}{\ell}\right)^{3}
           & r<\ell\\
           0 & r\geq\ell
           \end{cases}
        """
        r = np.array(np.abs(r), dtype=np.double)
        res = np.zeros_like(r)
        r_ll = r < self.len_scale
        r_low = r[r_ll]
        res[r_ll] = (
            1.0
            - 3.0 / 2.0 * r_low / self.len_scale
            + 1.0 / 2.0 * (r_low / self.len_scale) ** 3
        )
        return res


class Intersection(CovModel):
    r"""The Intersection covariance model.

    This model is derived from the relative intersection area of
    two d-dimensional spheres,
    where the middle points have a distance of :math:`r`
    and the diameters are given by :math:`\ell`.

    In 1D this is the Linear model, in 2D this is the Circular model
    and in 3D this is the Spherical model.

    Notes
    -----
    This model is given by the following correlation functions.

    In 1D:

    .. math::
       \mathrm{cor}(r) =
       \begin{cases}
       1-\frac{r}{\ell}
       & r<\ell\\
       0 & r\geq\ell
       \end{cases}

    In 2D:

    .. math::
       \mathrm{cor}(r) =
       \begin{cases}
       \frac{2}{\pi}\cdot\left(
       \cos^{-1}\left(\frac{r}{\ell}\right) -
       \frac{r}{\ell}\cdot\sqrt{1-\left(\frac{r}{\ell}\right)^{2}}
       \right)
       & r<\ell\\
       0 & r\geq\ell
       \end{cases}

    In 3D:

    .. math::
       \mathrm{cor}(r) =
       \begin{cases}
       1-\frac{3}{2}\cdot\frac{r}{\ell} +
       \frac{1}{2}\cdot\left(\frac{r}{\ell}\right)^{3}
       & r<\ell\\
       0 & r\geq\ell
       \end{cases}

    """

    def correlation(self, r):  # noqa: D102
        r = np.array(np.abs(r), dtype=np.double)
        res = np.zeros_like(r)
        r_ll = r < self.len_scale
        r_low = r[r_ll]
        if self.dim == 1:
            res[r_ll] = 1.0 - r_low / self.len_scale
        elif self.dim == 2:
            res[r_ll] = (
                2
                / np.pi
                * (
                    np.arccos(r_low / self.len_scale)
                    - r_low
                    / self.len_scale
                    * np.sqrt(1 - (r_low / self.len_scale) ** 2)
                )
            )
        else:
            res[r_ll] = (
                1.0
                - 3.0 / 2.0 * r_low / self.len_scale
                + 1.0 / 2.0 * (r_low / self.len_scale) ** 3
            )
        return res

    def spectrum(self, k):  # noqa: D102
        k = np.array(k, dtype=np.double)
        res = np.empty_like(k)
        kl = k * self.len_scale
        kl_gz = kl > 0
        # for k=0 we calculate the limit by hand
        if self.dim == 1:
            res[kl_gz] = (1.0 - np.cos(kl[kl_gz])) / (
                np.pi * k[kl_gz] * kl[kl_gz]
            )
            res[np.logical_not(kl_gz)] = self.len_scale / 2.0 / np.pi
        elif self.dim == 2:
            res[kl_gz] = sps.j1(kl[kl_gz] / 2.0) ** 2 / np.pi / k[kl_gz] ** 2
            res[np.logical_not(kl_gz)] = self.len_scale ** 2 / 16.0 / np.pi
        else:
            res[kl_gz] = -(
                12 * kl[kl_gz] * np.sin(kl[kl_gz])
                + (12 - 3 * kl[kl_gz] ** 2) * np.cos(kl[kl_gz])
                - 3 * kl[kl_gz] ** 2
                - 12
            ) / (2 * np.pi ** 2 * kl[kl_gz] ** 3 * k[kl_gz] ** 3)
            res[np.logical_not(kl_gz)] = (
                self.len_scale ** 3 / 48.0 / np.pi ** 2
            )
        return res * self.var



# -*- coding: utf-8 -*-
"""
GStools subpackage providing a class for standard spatial random fields.

.. currentmodule:: gstools.field.srf

The following classes are provided

.. autosummary::
   SRF
"""
# pylint: disable=C0103
from __future__ import division, absolute_import, print_function

import numpy as np
from gstools.field.generator import RandMeth, IncomprRandMeth
from gstools.field.tools import (
    check_mesh,
    make_isotropic,
    unrotate_mesh,
    reshape_axis_from_struct_to_unstruct,
    reshape_field_from_unstruct_to_struct,
)
from gstools.field.base import Field
from gstools.tools.geometric import pos2xyz, xyz2pos
from gstools.field.upscaling import var_coarse_graining, var_no_scaling
from gstools.field.condition import ordinary, simple
from gstools.krige.tools import set_condition

__all__ = ["SRF"]

GENERATOR = {
    "RandMeth": RandMeth,
    "IncomprRandMeth": IncomprRandMeth,
    "VectorField": IncomprRandMeth,
    "VelocityField": IncomprRandMeth,
}
UPSCALING = {
    "coarse_graining": var_coarse_graining,
    "no_scaling": var_no_scaling,
}
CONDITION = {"ordinary": ordinary, "simple": simple}


class SRF(Field):
    """A class to generate spatial random fields (SRF).

    Parameters
    ----------
    model : :any:`CovModel`
        Covariance Model of the spatial random field.
    mean : :class:`float`, optional
        mean value of the SRF
    upscaling : :class:`str`, optional
        Method to be used for upscaling the variance at each point
        depending on the related element volume.
        See the ``point_volumes`` keyword in the :any:`SRF.__call__` routine.
        At the moment, the following upscaling methods are provided:

            * "no_scaling" : No upscaling is applied to the variance.
              See: :any:`var_no_scaling`
            * "coarse_graining" : A volume depended variance is
              calculated by the upscaling technique coarse graining.
              See: :any:`var_coarse_graining`

        Default: "no_scaling"
    generator : :class:`str`, optional
        Name of the field generator to be used.
        At the moment, the following generators are provided:

            * "RandMeth" : The Randomization Method.
              See: :any:`RandMeth`
            * "IncomprRandMeth" : The incompressible Randomization Method.
              This is the original algorithm proposed by Kraichnan 1970
              See: :any:`IncomprRandMeth`
            * "VectorField" : an alias for "IncomprRandMeth"
            * "VelocityField" : an alias for "IncomprRandMeth"

        Default: "RandMeth"
    **generator_kwargs
        Keyword arguments that are forwarded to the generator in use.
        Have a look at the provided generators for further information.
    """

    def __init__(
        self,
        model,
        mean=0.0,
        upscaling="no_scaling",
        generator="RandMeth",
        **generator_kwargs
    ):
        super(SRF, self).__init__(model, mean)
        # initialize private attributes
        self._generator = None
        self._upscaling = None
        self._upscaling_func = None
        # condition related
        self._cond_pos = None
        self._cond_val = None
        self._krige_type = None
        # initialize attributes
        self.raw_field = None
        self.krige_field = None
        self.err_field = None
        self.krige_var = None
        self.set_generator(generator, **generator_kwargs)
        self.upscaling = upscaling
        if self._value_type is None:
            raise ValueError(
                "Unknown field value type, "
                + "specify 'scalar' or 'vector' before calling SRF."
            )

    def __call__(
        self, pos, seed=np.nan, point_volumes=0.0, mesh_type="unstructured"
    ):
        """Generate the spatial random field.

        The field is saved as `self.field` and is also returned.

        Parameters
        ----------
        pos : :class:`list`
            the position tuple, containing main direction and transversal
            directions
        seed : :class:`int`, optional
            seed for RNG for reseting. Default: keep seed from generator
        point_volumes : :class:`float` or :class:`numpy.ndarray`
            If your evaluation points for the field are coming from a mesh,
            they are probably representing a certain element volume.
            This volume can be passed by `point_volumes` to apply the
            given variance upscaling. If `point_volumes` is ``0`` nothing
            is changed. Default: ``0``
        mesh_type : :class:`str`
            'structured' / 'unstructured'

        Returns
        -------
        field : :class:`numpy.ndarray`
            the SRF
        """
        # internal conversation
        x, y, z = pos2xyz(pos, max_dim=self.model.dim)
        self.pos = xyz2pos(x, y, z)
        self.mesh_type = mesh_type
        # update the model/seed in the generator if any changes were made
        self.generator.update(self.model, seed)
        # format the positional arguments of the mesh
        check_mesh(self.model.dim, x, y, z, mesh_type)
        mesh_type_changed = False
        if self.model.do_rotation:
            if mesh_type == "structured":
                mesh_type_changed = True
                mesh_type_old = mesh_type
                mesh_type = "unstructured"
                x, y, z, axis_lens = reshape_axis_from_struct_to_unstruct(
                    self.model.dim, x, y, z
                )
            x, y, z = unrotate_mesh(self.model.dim, self.model.angles, x, y, z)
        y, z = make_isotropic(self.model.dim, self.model.anis, y, z)

        # generate the field
        self.raw_field = self.generator.__call__(x, y, z, mesh_type)

        # reshape field if we got an unstructured mesh
        if mesh_type_changed:
            mesh_type = mesh_type_old
            self.raw_field = reshape_field_from_unstruct_to_struct(
                self.model.dim, self.raw_field, axis_lens
            )

        # apply given conditions to the field
        if self.condition:
            (
                cond_field,
                krige_field,
                err_field,
                krigevar,
                info,
            ) = self.cond_func(self)
            # store everything in the class
            self.field = cond_field
            self.krige_field = krige_field
            self.err_field = err_field
            self.krige_var = krigevar
            if "mean" in info:  # ordinary krging estimates mean
                self.mean = info["mean"]
        else:
            self.field = self.raw_field + self.mean

        # upscaled variance
        if not np.isscalar(point_volumes) or not np.isclose(point_volumes, 0):
            scaled_var = self.upscaling_func(self.model, point_volumes)
            self.field -= self.mean
            self.field *= np.sqrt(scaled_var / self.model.sill)
            self.field += self.mean

        return self.field

    def set_condition(
        self, cond_pos=None, cond_val=None, krige_type="ordinary"
    ):
        """Condition a given spatial random field with measurements.

        Parameters
        ----------
        cond_pos : :class:`list`
            the position tuple of the conditions
        cond_val : :class:`numpy.ndarray`
            the values of the conditions
        krige_type : :class:`str`, optional
            Used kriging type for conditioning.
            Either 'ordinary' or 'simple'.
            Default: 'ordinary'

        Notes
        -----
        When using "ordinary" as ``krige_type``, the ``mean`` attribute of the
        spatial random field will be overwritten with the estimated mean.
        """
        if cond_pos is not None:
            self._cond_pos, self._cond_val = set_condition(
                cond_pos, cond_val, self.model.dim
            )
        else:
            self._cond_pos = self._cond_val = None
        self._krige_type = krige_type
        if krige_type not in CONDITION:
            raise ValueError(
                "gstools.SRF: Unknown kriging method: " + krige_type
            )

    def del_condition(self):
        """Delete Conditions."""
        self._cond_pos = None
        self._cond_val = None
        self._krige_type = None

    @property
    def cond_pos(self):
        """:class:`list`: The position tuple of the conditions."""
        return self._cond_pos

    @property
    def cond_val(self):
        """:class:`list`: The values of the conditions."""
        return self._cond_val

    @property
    def condition(self):
        """:any:`bool`: State if conditions ar given."""
        return self._cond_pos is not None

    def cond_func(self, *args, **kwargs):
        """Conditioning method applied to the field."""
        if self.condition:
            return CONDITION[self._krige_type](*args, **kwargs)
        return None

    def upscaling_func(self, *args, **kwargs):
        """Upscaling method applied to the field variance."""
        return self._upscaling_func(*args, **kwargs)

    def set_generator(self, generator, **generator_kwargs):
        """Set the generator for the field.

        Parameters
        ----------
        generator : :class:`str`, optional
            Name of the generator to use for field generation.
            Default: "RandMeth"
        **generator_kwargs
            keyword arguments that are forwarded to the generator in use.
        """
        if generator in GENERATOR:
            gen = GENERATOR[generator]
            self._generator = gen(self.model, **generator_kwargs)
            self._value_type = self._generator.value_type
        else:
            raise ValueError("gstools.SRF: Unknown generator: " + generator)

    @property
    def generator(self):
        """:any:`callable`: The generator of the field.

        Default: :any:`RandMeth`
        """
        return self._generator

    @property
    def upscaling(self):  # pragma: no cover
        """:class:`str`: Name of the upscaling method.

        See the ``point_volumes`` keyword in the :any:`SRF.__call__` routine.
        Default: "no_scaling"
        """
        return self._upscaling

    @upscaling.setter
    def upscaling(self, upscaling):
        if upscaling in UPSCALING:
            self._upscaling = upscaling
            self._upscaling_func = UPSCALING[upscaling]
        else:
            raise ValueError(
                "gstools.SRF: Unknown upscaling method: " + upscaling
            )

    def __repr__(self):
        """Return String representation."""
        return "SRF(model={0}, mean={1}, generator={2}".format(
            self.model, self.mean, self.generator
        )


if __name__ == "__main__":  # pragma: no cover
    import doctest

    doctest.testmod()



# -*- coding: utf-8 -*-
"""
GStools subpackage providing miscellaneous tools.

.. currentmodule:: gstools.tools

Export
^^^^^^

.. autosummary::
   to_vtk_structured
   vtk_export_structured
   to_vtk_unstructured
   vtk_export_unstructured
   to_vtk
   vtk_export

Special functions
^^^^^^^^^^^^^^^^^

.. autosummary::
   inc_gamma
   exp_int
   inc_beta
   tplstable_cor

Geometric
^^^^^^^^^

.. autosummary::
   xyz2pos
   pos2xyz
   r3d_x
   r3d_y
   r3d_z

----
"""
from __future__ import absolute_import

from gstools.tools.export import (
    vtk_export_structured,
    vtk_export_unstructured,
    vtk_export,
)

from gstools.tools.special import inc_gamma, exp_int, inc_beta, tplstable_cor

from gstools.tools.geometric import r3d_x, r3d_y, r3d_z, xyz2pos, pos2xyz

__all__ = [
    "vtk_export_structured",
    "vtk_export_unstructured",
    "vtk_export",
    "inc_gamma",
    "exp_int",
    "inc_beta",
    "tplstable_cor",
    "xyz2pos",
    "pos2xyz",
    "r3d_x",
    "r3d_y",
    "r3d_z",
]



# -*- coding: utf-8 -*-
"""
This is a unittest of the variogram module.
"""
from __future__ import division, absolute_import, print_function

import sys
import unittest
import numpy as np
from gstools import variogram

PY3 = sys.version_info[0] == 3
# in python3 "long" was replaced with "int"
if PY3:
    LONGTYPE = int
else:
    LONGTYPE = long


class TestVariogramstructured(unittest.TestCase):
    def setUp(self):
        pass

    def test_doubles(self):
        z = np.array(
            (41.2, 40.2, 39.7, 39.2, 40.1, 38.3, 39.1, 40.0, 41.1, 40.3),
            dtype=np.double,
        )
        try:
            gamma = variogram.vario_estimate_structured(z)
            self.assertAlmostEqual(gamma[1], 0.4917, places=4)
        except NotImplementedError as e:
            pass

    def test_ints(self):
        z = np.array((10, 20, 30, 40), dtype=int)
        try:
            gamma = variogram.vario_estimate_structured(z)
            self.assertAlmostEqual(gamma[1], 50.0, places=4)
        except NotImplementedError as e:
            pass

    def test_longs(self):
        z = np.array((10, 20, 30, 40), dtype=LONGTYPE)
        try:
            gamma = variogram.vario_estimate_structured(z)
            self.assertAlmostEqual(gamma[1], 50.0, places=4)
        except NotImplementedError as e:
            pass

    def test_np_int(self):
        z = np.array((10, 20, 30, 40), dtype=np.int)
        try:
            gamma = variogram.vario_estimate_structured(z)
            self.assertAlmostEqual(gamma[1], 50.0, places=4)
        except NotImplementedError as e:
            pass

    def test_mixed(self):
        z = np.array(
            (41.2, 40.2, 39.7, 39.2, 40.1, 38.3, 39.1, 40.0, 41.1, 40.3),
            dtype=np.double,
        )
        try:
            gamma = variogram.vario_estimate_structured(z)
            self.assertAlmostEqual(gamma[1], 0.4917, places=4)

            z = np.array((10, 20, 30, 40), dtype=LONGTYPE)

            gamma = variogram.vario_estimate_structured(z)
            self.assertAlmostEqual(gamma[1], 50.0, places=4)

            z = np.array((10, 20, 30, 40), dtype=LONGTYPE)
            gamma = variogram.vario_estimate_structured(z)
            self.assertAlmostEqual(gamma[1], 50.0, places=4)
        except NotImplementedError as e:
            pass

    def test_list(self):
        z = [41.2, 40.2, 39.7, 39.2, 40.1, 38.3, 39.1, 40.0, 41.1, 40.3]
        try:
            gamma = variogram.vario_estimate_structured(z)
            self.assertAlmostEqual(gamma[1], 0.4917, places=4)
        except NotImplementedError as e:
            pass

    def test_1d(self):
        # literature values
        z = np.array(
            (41.2, 40.2, 39.7, 39.2, 40.1, 38.3, 39.1, 40.0, 41.1, 40.3),
            dtype=np.double,
        )
        try:
            gamma = variogram.vario_estimate_structured(z)
            self.assertAlmostEqual(gamma[0], 0.0000, places=4)
            self.assertAlmostEqual(gamma[1], 0.4917, places=4)
            self.assertAlmostEqual(gamma[2], 0.7625, places=4)
        except NotImplementedError as e:
            pass

    def test_masked_1d(self):
        # literature values
        z = np.array(
            (41.2, 40.2, 39.7, 39.2, 40.1, 38.3, 39.1, 40.0, 41.1, 40.3),
            dtype=np.double,
        )
        z_ma = np.ma.masked_array(z, mask=[0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        try:
            gamma = variogram.vario_estimate_structured(z_ma)
            self.assertAlmostEqual(gamma[0], 0.0000, places=4)
            self.assertAlmostEqual(gamma[1], 0.4917, places=4)
            self.assertAlmostEqual(gamma[2], 0.7625, places=4)
            z_ma = np.ma.masked_array(z, mask=[1, 0, 0, 0, 0, 0, 0, 0, 0, 0])
            gamma = variogram.vario_estimate_structured(z_ma)
            self.assertAlmostEqual(gamma[0], 0.0000, places=4)
            self.assertAlmostEqual(gamma[1], 0.4906, places=4)
            self.assertAlmostEqual(gamma[2], 0.7107, places=4)
        except NotImplementedError as e:
            pass

    def test_masked_2d(self):
        rng = np.random.RandomState(1479373475)
        field = rng.rand(80, 60)
        mask = np.zeros_like(field)
        field_ma = np.ma.masked_array(field, mask=mask)

        try:
            gamma_x = variogram.vario_estimate_structured(
                field_ma, direction="x"
            )
            gamma_y = variogram.vario_estimate_structured(
                field_ma, direction="y"
            )

            var = 1.0 / 12.0
            self.assertAlmostEqual(gamma_x[0], 0.0, places=2)
            self.assertAlmostEqual(gamma_x[len(gamma_x) // 2], var, places=2)
            self.assertAlmostEqual(gamma_x[-1], var, places=2)
            self.assertAlmostEqual(gamma_y[0], 0.0, places=2)
            self.assertAlmostEqual(gamma_y[len(gamma_y) // 2], var, places=2)
            self.assertAlmostEqual(gamma_y[-1], var, places=2)

            mask = np.zeros_like(field)
            mask[0, 0] = 1
            field = np.ma.masked_array(field, mask=mask)
            gamma_x = variogram.vario_estimate_structured(
                field_ma, direction="x"
            )
            gamma_y = variogram.vario_estimate_structured(
                field_ma, direction="y"
            )
            self.assertAlmostEqual(gamma_x[0], 0.0, places=2)
            self.assertAlmostEqual(gamma_y[0], 0.0, places=2)
        except NotImplementedError as e:
            pass

    def test_masked_3d(self):
        rng = np.random.RandomState(1479373475)
        field = rng.rand(30, 30, 30)
        mask = np.zeros_like(field)
        field_ma = np.ma.masked_array(field, mask=mask)

        try:
            gamma_x = variogram.vario_estimate_structured(
                field_ma, direction="x"
            )
            gamma_y = variogram.vario_estimate_structured(
                field_ma, direction="y"
            )
            gamma_z = variogram.vario_estimate_structured(
                field_ma, direction="z"
            )

            var = 1.0 / 12.0
            self.assertAlmostEqual(gamma_x[0], 0.0, places=2)
            self.assertAlmostEqual(gamma_x[len(gamma_x) // 2], var, places=2)
            self.assertAlmostEqual(gamma_x[-1], var, places=2)
            self.assertAlmostEqual(gamma_y[0], 0.0, places=2)
            self.assertAlmostEqual(gamma_y[len(gamma_y) // 2], var, places=2)
            self.assertAlmostEqual(gamma_y[-1], var, places=2)
            self.assertAlmostEqual(gamma_z[0], 0.0, places=2)
            self.assertAlmostEqual(gamma_z[len(gamma_y) // 2], var, places=2)
            self.assertAlmostEqual(gamma_z[-1], var, places=2)

            mask = np.zeros_like(field)
            mask[0, 0, 0] = 1
            field = np.ma.masked_array(field, mask=mask)
            gamma_x = variogram.vario_estimate_structured(
                field_ma, direction="x"
            )
            gamma_y = variogram.vario_estimate_structured(
                field_ma, direction="y"
            )
            gamma_z = variogram.vario_estimate_structured(
                field_ma, direction="z"
            )
            self.assertAlmostEqual(gamma_x[0], 0.0, places=2)
            self.assertAlmostEqual(gamma_y[0], 0.0, places=2)
            self.assertAlmostEqual(gamma_z[0], 0.0, places=2)
        except NotImplementedError as e:
            pass

    def test_uncorrelated_2d(self):
        x = np.linspace(0.0, 100.0, 80)
        y = np.linspace(0.0, 100.0, 60)

        rng = np.random.RandomState(1479373475)
        field = rng.rand(len(x), len(y))

        try:
            gamma_x = variogram.vario_estimate_structured(field, direction="x")
            gamma_y = variogram.vario_estimate_structured(field, direction="y")

            var = 1.0 / 12.0
            self.assertAlmostEqual(gamma_x[0], 0.0, places=2)
            self.assertAlmostEqual(gamma_x[len(gamma_x) // 2], var, places=2)
            self.assertAlmostEqual(gamma_x[-1], var, places=2)
            self.assertAlmostEqual(gamma_y[0], 0.0, places=2)
            self.assertAlmostEqual(gamma_y[len(gamma_y) // 2], var, places=2)
            self.assertAlmostEqual(gamma_y[-1], var, places=2)
        except NotImplementedError as e:
            pass

    def test_uncorrelated_3d(self):
        x = np.linspace(0.0, 100.0, 30)
        y = np.linspace(0.0, 100.0, 30)
        z = np.linspace(0.0, 100.0, 30)

        rng = np.random.RandomState(1479373475)
        field = rng.rand(len(x), len(y), len(z))

        try:
            gamma = variogram.vario_estimate_structured(field, "x")
            gamma = variogram.vario_estimate_structured(field, "y")
            gamma = variogram.vario_estimate_structured(field, "z")

            var = 1.0 / 12.0
            self.assertAlmostEqual(gamma[0], 0.0, places=2)
            self.assertAlmostEqual(gamma[len(gamma) // 2], var, places=2)
            self.assertAlmostEqual(gamma[-1], var, places=2)
        except NotImplementedError as e:
            pass

    def test_directions_2d(self):
        x = np.linspace(0.0, 20.0, 100)
        y = np.linspace(0.0, 15.0, 80)
        rng = np.random.RandomState(1479373475)
        x_rand = rng.rand(len(x))
        y_rand = rng.rand(len(y))
        # random values repeated along y-axis
        field_x = np.tile(x_rand, (len(y), 1)).T
        # random values repeated along x-axis
        field_y = np.tile(y_rand, (len(x), 1))

        try:
            gamma_x_x = variogram.vario_estimate_structured(
                field_x, direction="x"
            )
            gamma_x_y = variogram.vario_estimate_structured(
                field_x, direction="y"
            )

            gamma_y_x = variogram.vario_estimate_structured(
                field_y, direction="x"
            )
            gamma_y_y = variogram.vario_estimate_structured(
                field_y, direction="y"
            )

            self.assertAlmostEqual(gamma_x_y[1], 0.0)
            self.assertAlmostEqual(gamma_x_y[len(gamma_x_y) // 2], 0.0)
            self.assertAlmostEqual(gamma_x_y[-1], 0.0)
            self.assertAlmostEqual(gamma_y_x[1], 0.0)
            self.assertAlmostEqual(gamma_y_x[len(gamma_x_y) // 2], 0.0)
            self.assertAlmostEqual(gamma_y_x[-1], 0.0)
        except NotImplementedError as e:
            pass

    def test_directions_3d(self):
        x = np.linspace(0.0, 10.0, 20)
        y = np.linspace(0.0, 15.0, 25)
        z = np.linspace(0.0, 20.0, 30)
        rng = np.random.RandomState(1479373475)
        x_rand = rng.rand(len(x))
        y_rand = rng.rand(len(y))
        z_rand = rng.rand(len(z))

        field_x = np.tile(x_rand.reshape((len(x), 1, 1)), (1, len(y), len(z)))
        field_y = np.tile(y_rand.reshape((1, len(y), 1)), (len(x), 1, len(z)))
        field_z = np.tile(z_rand.reshape((1, 1, len(z))), (len(x), len(y), 1))

        try:
            gamma_x_x = variogram.vario_estimate_structured(
                field_x, direction="x"
            )
            gamma_x_y = variogram.vario_estimate_structured(
                field_x, direction="y"
            )
            gamma_x_z = variogram.vario_estimate_structured(
                field_x, direction="z"
            )

            gamma_y_x = variogram.vario_estimate_structured(
                field_y, direction="x"
            )
            gamma_y_y = variogram.vario_estimate_structured(
                field_y, direction="y"
            )
            gamma_y_z = variogram.vario_estimate_structured(
                field_y, direction="z"
            )

            gamma_z_x = variogram.vario_estimate_structured(
                field_z, direction="x"
            )
            gamma_z_y = variogram.vario_estimate_structured(
                field_z, direction="y"
            )
            gamma_z_z = variogram.vario_estimate_structured(
                field_z, direction="z"
            )
            self.assertAlmostEqual(gamma_x_y[1], 0.0)
            self.assertAlmostEqual(gamma_x_y[len(gamma_x_y) // 2], 0.0)
            self.assertAlmostEqual(gamma_x_y[-1], 0.0)
            self.assertAlmostEqual(gamma_x_z[1], 0.0)
            self.assertAlmostEqual(gamma_x_z[len(gamma_x_y) // 2], 0.0)
            self.assertAlmostEqual(gamma_x_z[-1], 0.0)
            self.assertAlmostEqual(gamma_y_x[1], 0.0)
            self.assertAlmostEqual(gamma_y_x[len(gamma_x_y) // 2], 0.0)
            self.assertAlmostEqual(gamma_y_x[-1], 0.0)
            self.assertAlmostEqual(gamma_y_z[1], 0.0)
            self.assertAlmostEqual(gamma_y_z[len(gamma_x_y) // 2], 0.0)
            self.assertAlmostEqual(gamma_y_z[-1], 0.0)
            self.assertAlmostEqual(gamma_z_x[1], 0.0)
            self.assertAlmostEqual(gamma_z_x[len(gamma_x_y) // 2], 0.0)
            self.assertAlmostEqual(gamma_z_x[-1], 0.0)
            self.assertAlmostEqual(gamma_z_y[1], 0.0)
            self.assertAlmostEqual(gamma_z_y[len(gamma_x_y) // 2], 0.0)
            self.assertAlmostEqual(gamma_z_y[-1], 0.0)
        except NotImplementedError as e:
            pass

    def test_exceptions(self):
        x = np.linspace(0.0, 10.0, 20)
        rng = np.random.RandomState(1479373475)
        x_rand = rng.rand(len(x))
        self.assertRaises(
            ValueError, variogram.vario_estimate_structured, x, "a"
        )


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
GStools subpackage providing geometric tools.

.. currentmodule:: gstools.tools.geometric

The following functions are provided

.. autosummary::
   r3d_x
   r3d_y
   r3d_z
   pos2xyz
   xyz2pos
"""
# pylint: disable=C0103
from __future__ import print_function, division, absolute_import

import numpy as np

__all__ = ["r3d_x", "r3d_y", "r3d_z", "pos2xyz", "xyz2pos"]


# Geometric functions #########################################################


def r3d_x(theta):
    """Rotation matrix about x axis.

    Parameters
    ----------
    theta : :class:`float`
        Rotation angle

    Returns
    -------
        :class:`numpy.ndarray`
            Rotation matrix.
    """
    sin = np.sin(theta)
    cos = np.cos(theta)
    return np.array(((1.0, +0.0, +0.0), (0.0, cos, -sin), (0.0, sin, cos)))


def r3d_y(theta):
    """Rotation matrix about y axis.

    Parameters
    ----------
    theta : :class:`float`
        Rotation angle

    Returns
    -------
        :class:`numpy.ndarray`
            Rotation matrix.
    """
    sin = np.sin(theta)
    cos = np.cos(theta)
    return np.array(((+cos, 0.0, sin), (+0.0, 1.0, +0.0), (-sin, 0.0, cos)))


def r3d_z(theta):
    """Rotation matrix about z axis.

    Parameters
    ----------
    theta : :class:`float`
        Rotation angle

    Returns
    -------
        :class:`numpy.ndarray`
            Rotation matrix.
    """
    sin = np.sin(theta)
    cos = np.cos(theta)
    return np.array(((cos, -sin, 0.0), (sin, +cos, 0.0), (+0.0, +0.0, 1.0)))


# conversion ##################################################################


def pos2xyz(pos, dtype=None, calc_dim=False, max_dim=3):
    """Convert postional arguments to x, y, z.

    Parameters
    ----------
    pos : :any:`iterable`
        the position tuple, containing main direction and transversal
        directions
    dtype : data-type, optional
        The desired data-type for the array.
        If not given, then the type will be determined as the minimum type
        required to hold the objects in the sequence. Default: None
    calc_dim : :class:`bool`, optional
        State if the dimension should be returned. Default: False
    max_dim : :class:`int`, optional
        Cut of information above the given dimension. Default: 3

    Returns
    -------
    x : :class:`numpy.ndarray`
        first components of position vectors
    y : :class:`numpy.ndarray` or None
        analog to x
    z : :class:`numpy.ndarray` or None
        analog to x
    dim : :class:`int`, optional
        dimension (only if calc_dim is True)

    Notes
    -----
    If len(pos) > 3, everything after pos[2] will be ignored.
    """
    if max_dim == 1:  # sanity check
        pos = np.array(pos, ndmin=2)
    x = np.array(pos[0], dtype=dtype).reshape(-1)
    dim = 1
    y = z = None
    if len(pos) > 1 and max_dim > 1:
        dim = 2
        y = np.array(pos[1], dtype=dtype).reshape(-1)
    if len(pos) > 2 and max_dim > 2:
        dim = 3
        z = np.array(pos[2], dtype=dtype).reshape(-1)
    if calc_dim:
        return x, y, z, dim
    return x, y, z


def xyz2pos(x, y=None, z=None, dtype=None, max_dim=3):
    """Convert x, y, z to postional arguments.

    Parameters
    ----------
    x : :class:`numpy.ndarray`
        grid axis in x-direction if structured, or first components of
        position vectors if unstructured
    y : :class:`numpy.ndarray`, optional
        analog to x
    z : :class:`numpy.ndarray`, optional
        analog to x
    dtype : data-type, optional
        The desired data-type for the array.
        If not given, then the type will be determined as the minimum type
        required to hold the objects in the sequence. Default: None
    max_dim : :class:`int`, optional
        Cut of information above the given dimension. Default: 3

    Returns
    -------
    pos : :class:`numpy.ndarray`
        the position tuple
    """
    if y is None and z is not None:
        raise ValueError("gstools.tools.xyz2pos: if z is given, y is needed!")
    pos = []
    pos.append(np.array(x, dtype=dtype).reshape(-1))
    if y is not None and max_dim > 1:
        pos.append(np.array(y, dtype=dtype).reshape(-1))
    if z is not None and max_dim > 2:
        pos.append(np.array(z, dtype=dtype).reshape(-1))
    return tuple(pos)



# -*- coding: utf-8 -*-
"""
GStools subpackage providing a class for simple kriging.

.. currentmodule:: gstools.krige.simple

The following classes are provided

.. autosummary::
   Simple
"""
# pylint: disable=C0103
from __future__ import division, absolute_import, print_function


import numpy as np
from scipy.linalg import inv
from scipy.spatial.distance import cdist

from gstools.field.tools import (
    check_mesh,
    make_isotropic,
    unrotate_mesh,
    reshape_axis_from_struct_to_unstruct,
    reshape_field_from_unstruct_to_struct,
)
from gstools.field.base import Field
from gstools.tools.geometric import pos2xyz, xyz2pos
from gstools.krige.krigesum import krigesum
from gstools.krige.tools import set_condition

__all__ = ["Simple"]


class Simple(Field):
    """
    A class for simple kriging.

    Parameters
    ----------
    model : :any:`CovModel`
        Covariance Model used for kriging.
    mean : :class:`float`, optional
        mean value of the kriging field
    cond_pos : :class:`list`
        tuple, containing the given condition positions (x, [y, z])
    cond_val : :class:`numpy.ndarray`
        the values of the conditions
    """

    def __init__(self, model, mean, cond_pos, cond_val):
        super(Simple, self).__init__(model, mean)
        self.krige_var = None
        # initialize private attributes
        self._value_type = "scalar"
        self._cond_pos = None
        self._cond_val = None
        self.set_condition(cond_pos, cond_val)

    def __call__(self, pos, mesh_type="unstructured"):
        """
        Generate the simple kriging field.

        The field is saved as `self.field` and is also returned.

        Parameters
        ----------
        pos : :class:`list`
            the position tuple, containing main direction and transversal
            directions (x, [y, z])
        mesh_type : :class:`str`
            'structured' / 'unstructured'

        Returns
        -------
        field : :class:`numpy.ndarray`
            the kriged field
        krige_var : :class:`numpy.ndarray`
            the kriging error variance
        """
        # internal conversation
        x, y, z = pos2xyz(pos, dtype=np.double, max_dim=self.model.dim)
        c_x, c_y, c_z = pos2xyz(
            self.cond_pos, dtype=np.double, max_dim=self.model.dim
        )
        self.pos = xyz2pos(x, y, z)
        self.mesh_type = mesh_type
        # format the positional arguments of the mesh
        check_mesh(self.model.dim, x, y, z, mesh_type)
        mesh_type_changed = False
        if mesh_type == "structured":
            mesh_type_changed = True
            mesh_type_old = mesh_type
            mesh_type = "unstructured"
            x, y, z, axis_lens = reshape_axis_from_struct_to_unstruct(
                self.model.dim, x, y, z
            )
        if self.model.do_rotation:
            x, y, z = unrotate_mesh(self.model.dim, self.model.angles, x, y, z)
            c_x, c_y, c_z = unrotate_mesh(
                self.model.dim, self.model.angles, c_x, c_y, c_z
            )
        y, z = make_isotropic(self.model.dim, self.model.anis, y, z)
        c_y, c_z = make_isotropic(self.model.dim, self.model.anis, c_y, c_z)

        # set condtions to zero mean
        cond = self.cond_val - self.mean
        krig_mat = inv(self._get_cov_mat((c_x, c_y, c_z), (c_x, c_y, c_z)))
        krig_vecs = self._get_cov_mat((c_x, c_y, c_z), (x, y, z))
        # generate the kriged field
        field, krige_var = krigesum(krig_mat, krig_vecs, cond)

        # reshape field if we got an unstructured mesh
        if mesh_type_changed:
            mesh_type = mesh_type_old
            field = reshape_field_from_unstruct_to_struct(
                self.model.dim, field, axis_lens
            )
            krige_var = reshape_field_from_unstruct_to_struct(
                self.model.dim, krige_var, axis_lens
            )
        # calculate the kriging error
        self.krige_var = self.model.sill - krige_var
        # add the given mean
        self.field = field + self.mean
        return self.field, self.krige_var

    def _get_cov_mat(self, pos1, pos2):
        return self.model.cov_nugget(
            cdist(
                np.column_stack(pos1[: self.model.dim]),
                np.column_stack(pos2[: self.model.dim]),
            )
        )

    def set_condition(self, cond_pos, cond_val):
        """Set the conditions for kriging.

        Parameters
        ----------
        cond_pos : :class:`list`
            the position tuple of the conditions (x, [y, z])
        cond_val : :class:`numpy.ndarray`
            the values of the conditions
        """
        self._cond_pos, self._cond_val = set_condition(
            cond_pos, cond_val, self.model.dim
        )

    @property
    def cond_pos(self):
        """:class:`list`: The position tuple of the conditions."""
        return self._cond_pos

    @property
    def cond_val(self):
        """:class:`list`: The values of the conditions."""
        return self._cond_val

    def __repr__(self):
        """Return String representation."""
        return "Simple(model={0}, mean={1}, cond_pos={2}, cond_val={3}".format(
            self.model, self.mean, self.cond_pos, self.cond_val
        )


if __name__ == "__main__":  # pragma: no cover
    import doctest

    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Purpose
=======

GeoStatTools is a library providing geostatistical tools
for random field generation, conditioned field generation,
kriging and variogram estimation
based on a list of provided or even user-defined covariance models.

The following functionalities are directly provided on module-level.

Subpackages
===========

.. autosummary::
    covmodel
    field
    variogram
    krige
    random
    tools
    transform

Classes
=======

Spatial Random Field
^^^^^^^^^^^^^^^^^^^^
Class for random field generation

.. currentmodule:: gstools.field

.. autosummary::
   SRF

Covariance Base-Class
^^^^^^^^^^^^^^^^^^^^^
Class to construct user defined covariance models

.. currentmodule:: gstools.covmodel.base

.. autosummary::
   CovModel

Covariance Models
^^^^^^^^^^^^^^^^^

Standard Covariance Models
~~~~~~~~~~~~~~~~~~~~~~~~~~

.. currentmodule:: gstools.covmodel.models

.. autosummary::
   Gaussian
   Exponential
   Matern
   Rational
   Stable
   Linear
   Circular
   Spherical
   Intersection

Truncated Power Law Covariance Models
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. currentmodule:: gstools.covmodel.tpl_models
.. autosummary::
   TPLGaussian
   TPLExponential
   TPLStable

Functions
=========

VTK-Export
^^^^^^^^^^
Routines to export fields to the vtk format

.. currentmodule:: gstools.tools

.. autosummary::
   to_vtk
   vtk_export
   to_vtk_structured
   vtk_export_structured
   to_vtk_unstructured
   vtk_export_unstructured

variogram estimation
^^^^^^^^^^^^^^^^^^^^
Estimate the variogram of a given field

.. currentmodule:: gstools.variogram

.. autosummary::
   vario_estimate_structured
   vario_estimate_unstructured
"""

from __future__ import absolute_import

from gstools._version import __version__
from gstools import field, variogram, random, covmodel, tools, krige, transform
from gstools.field import SRF
from gstools.tools.export import (
    vtk_export_structured,
    vtk_export_unstructured,
    vtk_export,
)
from gstools.variogram import (
    vario_estimate_structured,
    vario_estimate_unstructured,
)
from gstools.covmodel import (
    CovModel,
    Gaussian,
    Exponential,
    Matern,
    Rational,
    Stable,
    Linear,
    Circular,
    Spherical,
    Intersection,
    TPLGaussian,
    TPLExponential,
    TPLStable,
)

__all__ = ["__version__"]
__all__ += ["covmodel", "field", "variogram", "krige", "random", "tools"]
__all__ += ["transform"]
__all__ += [
    "CovModel",
    "Gaussian",
    "Exponential",
    "Matern",
    "Rational",
    "Stable",
    "Linear",
    "Circular",
    "Spherical",
    "Intersection",
    "TPLGaussian",
    "TPLExponential",
    "TPLStable",
]

__all__ += ["vario_estimate_structured", "vario_estimate_unstructured"]

__all__ += [
    "SRF",
    "to_vtk_structured",
    "vtk_export_structured",
    "to_vtk_unstructured",
    "vtk_export_unstructured",
    "to_vtk",
    "vtk_export",
]



# -*- coding: utf-8 -*-
"""
GStools subpackage providing tools for the covariance-model.

.. currentmodule:: gstools.covmodel.tools

The following classes and functions are provided

.. autosummary::
   InitSubclassMeta
   rad_fac
   set_len_anis
   check_bounds
   inc_gamma
   exp_int
   inc_beta
"""
# pylint: disable=C0103
from __future__ import print_function, division, absolute_import

import numpy as np
from scipy import special as sps

__all__ = ["InitSubclassMeta", "rad_fac", "set_len_anis", "check_bounds"]


# __init_subclass__ hack ######################################################

if hasattr(object, "__init_subclass__"):
    InitSubclassMeta = type
else:

    class InitSubclassMeta(type):  # pragma: no cover
        """Metaclass that implements PEP 487 protocol.

        Notes
        -----
        See :
            https://www.python.org/dev/peps/pep-0487

        taken from :
            https://github.com/graphql-python/graphene/blob/master/graphene/pyutils/init_subclass.py
        """

        def __new__(cls, name, bases, ns, **kwargs):
            """Create a new subclass."""
            __init_subclass__ = ns.pop("__init_subclass__", None)
            if __init_subclass__:
                __init_subclass__ = classmethod(__init_subclass__)
                ns["__init_subclass__"] = __init_subclass__
            return super(InitSubclassMeta, cls).__new__(
                cls, name, bases, ns, **kwargs
            )

        def __init__(cls, name, bases, ns, **kwargs):
            super(InitSubclassMeta, cls).__init__(name, bases, ns)
            super_class = super(cls, cls)
            if hasattr(super_class, "__init_subclass__"):
                super_class.__init_subclass__.__func__(cls, **kwargs)


# Helping functions ###########################################################


def rad_fac(dim, r):
    """Volume element of the n-dimensional spherical coordinates.

    Given as a factor for integration of a radial-symmetric function.

    Parameters
    ----------
    dim : :class:`int`
        spatial dimension
    r : :class:`numpy.ndarray`
        Given radii.
    """
    if dim == 1:
        fac = 2.0
    elif dim == 2:
        fac = 2 * np.pi * r
    elif dim == 3:
        fac = 4 * np.pi * r ** 2
    else:  # general solution ( for the record :D )
        fac = (
            dim
            * r ** (dim - 1)
            * np.sqrt(np.pi) ** dim
            / sps.gamma(dim / 2.0 + 1.0)
        )
    return fac


def set_len_anis(dim, len_scale, anis):
    """Set the length scale and anisotropy factors for the given dimension.

    Parameters
    ----------
    dim : :class:`int`
        spatial dimension
    len_scale : :class:`float` or :class:`list`
        the length scale of the SRF in x direction or in x- (y-, z-) direction
    anis : :class:`float`/list
        the anisotropy of length scales along the y- and z-directions

    Returns
    -------
    len_scale : :class:`float`
        the main length scale of the SRF in x direction
    anis : :class:`float`/list, optional
        the anisotropy of length scales along the y- and z-directions

    Notes
    -----
    If ``len_scale`` is given as list, ``anis`` will be recalculated.
    """
    ls_tmp = np.atleast_1d(len_scale)[:dim]
    # use just one length scale (x-direction)
    out_len_scale = ls_tmp[0]
    # set the anisotropies in y- and z-direction according to the input
    if len(ls_tmp) == 1:
        out_anis = np.atleast_1d(anis)[: dim - 1]
        if len(out_anis) < dim - 1:
            # fill up the anisotropies with ones, such that len()==dim-1
            out_anis = np.pad(
                out_anis,
                (0, dim - len(out_anis) - 1),
                "constant",
                constant_values=1.0,
            )
    elif dim == 1:
        # there is no anisotropy in 1 dimension
        out_anis = np.empty(0)
    else:
        # fill up length-scales with main len_scale, such that len()==dim
        if len(ls_tmp) < dim:
            ls_tmp = np.pad(
                ls_tmp,
                (0, dim - len(ls_tmp)),
                "constant",
                constant_values=out_len_scale,
            )
        # if multiple length-scales are given, calculate the anisotropies
        out_anis = np.zeros(dim - 1, dtype=np.double)
        for i in range(1, dim):
            out_anis[i - 1] = ls_tmp[i] / ls_tmp[0]

    for ani in out_anis:
        if not ani > 0.0:
            raise ValueError(
                "anisotropy-ratios needs to be > 0, " + "got: " + str(out_anis)
            )
    return out_len_scale, out_anis


def set_angles(dim, angles):
    """Set the angles for the given dimension.

    Parameters
    ----------
    dim : :class:`int`
        spatial dimension (anything different from 1 and 2 is interpreted as 3)
    angles : :class:`float`/list
        the angles of the SRF

    Returns
    -------
    angles : :class:`float`
        the angles fitting to the dimension
    """
    if dim == 1:
        # no rotation in 1D
        out_angles = np.empty(0)
    elif dim == 2:
        # one rotation axis in 2D
        out_angles = np.atleast_1d(angles)[:1]
        # fill up the rotation angle array with zeros
        out_angles = np.pad(
            out_angles,
            (0, 1 - len(out_angles)),
            "constant",
            constant_values=0.0,
        )
    else:
        # three rotation axis in 3D
        out_angles = np.atleast_1d(angles)[:3]
        # fill up the rotation angle array with zeros
        out_angles = np.pad(
            out_angles,
            (0, 3 - len(out_angles)),
            "constant",
            constant_values=0.0,
        )
    return out_angles


def check_bounds(bounds):
    """
    Check if given bounds are valid.

    Parameters
    ----------
    bounds : list
        bound can contain 2 to 3 values:
            1. lower bound
                float
            2. upper bound
                float
            3. Interval type (optional)
                * "oo" : open - open
                * "oc" : open - close
                * "co" : close - open
                * "cc" : close - close
    """
    if len(bounds) not in (2, 3):
        return False
    if bounds[1] <= bounds[0]:
        return False
    if len(bounds) == 3 and bounds[2] not in ("oo", "oc", "co", "cc"):
        return False
    return True



# -*- coding: utf-8 -*-
"""
This is the unittest of the RNG class.
"""
from __future__ import division, absolute_import, print_function

import unittest
import numpy as np
from scipy.stats import skew, kurtosis
from scipy.stats import normaltest
from gstools import Gaussian, TPLStable
from gstools.random.rng import RNG


class TestRNG(unittest.TestCase):
    def setUp(self):
        self.seed = 19031977
        self.rng = RNG(self.seed)
        self.many_modes = 1000000
        self.few_modes = 100

    def test_rng_normal_consistency(self):
        rng = RNG(21021997)
        z1_refs = [-1.93013270, 0.46330478]
        z2_refs = [-0.25536086, 0.98298696]

        z1 = self.rng.random.normal(size=self.few_modes)
        z2 = self.rng.random.normal(size=self.few_modes)
        self.assertAlmostEqual(z1[0], z1_refs[0])
        self.assertAlmostEqual(z1[1], z1_refs[1])
        self.assertAlmostEqual(z2[0], z2_refs[0])
        self.assertAlmostEqual(z2[1], z2_refs[1])
        self.rng.seed = self.seed
        z1 = self.rng.random.normal(size=self.few_modes)
        z2 = self.rng.random.normal(size=self.few_modes)
        self.assertAlmostEqual(z1[0], z1_refs[0])
        self.assertAlmostEqual(z1[1], z1_refs[1])
        self.assertAlmostEqual(z2[0], z2_refs[0])
        self.assertAlmostEqual(z2[1], z2_refs[1])

    def test_sample_sphere_1d(self):
        dim = 1
        sphere_coord = self.rng.sample_sphere(dim, self.few_modes)
        self.assertEqual(sphere_coord.shape, (dim, self.few_modes))
        sphere_coord = self.rng.sample_sphere(dim, self.many_modes)
        self.assertAlmostEqual(np.mean(sphere_coord), 0.0, places=3)

    def test_sample_sphere_2d(self):
        dim = 2
        sphere_coord = self.rng.sample_sphere(dim, self.few_modes)
        np.testing.assert_allclose(
            np.ones(self.few_modes),
            sphere_coord[0, :] ** 2 + sphere_coord[1, :] ** 2,
        )
        sphere_coord = self.rng.sample_sphere(dim, self.many_modes)
        self.assertAlmostEqual(np.mean(sphere_coord), 0.0, places=3)

    def test_sample_sphere_3d(self):
        dim = 3
        sphere_coord = self.rng.sample_sphere(dim, self.few_modes)
        self.assertEqual(sphere_coord.shape, (dim, self.few_modes))
        np.testing.assert_allclose(
            np.ones(self.few_modes),
            sphere_coord[0, :] ** 2
            + sphere_coord[1, :] ** 2
            + sphere_coord[2, :] ** 2,
        )
        sphere_coord = self.rng.sample_sphere(dim, self.many_modes)
        self.assertAlmostEqual(np.mean(sphere_coord), 0.0, places=3)

    def test_sample_dist(self):
        model = Gaussian(dim=1, var=3.5, len_scale=8.0)
        pdf, cdf, ppf = model.dist_func
        rad = self.rng.sample_dist(
            size=self.few_modes, pdf=pdf, cdf=cdf, ppf=ppf, a=0
        )
        self.assertEqual(rad.shape[0], self.few_modes)

        model = Gaussian(dim=2, var=3.5, len_scale=8.0)
        pdf, cdf, ppf = model.dist_func
        rad = self.rng.sample_dist(
            size=self.few_modes, pdf=pdf, cdf=cdf, ppf=ppf, a=0
        )
        self.assertEqual(rad.shape[0], self.few_modes)

        model = Gaussian(dim=3, var=3.5, len_scale=8.0)
        pdf, cdf, ppf = model.dist_func
        rad = self.rng.sample_dist(
            size=self.few_modes, pdf=pdf, cdf=cdf, ppf=ppf, a=0
        )
        self.assertEqual(rad.shape[0], self.few_modes)

        # model = Gaussian(dim=2, var=3.5, len_scale=8.)
        # pdf, cdf, ppf = model.dist_func
        # rad = self.rng.sample_dist(
        #    size=self.many_modes, pdf=pdf, cdf=cdf, ppf=ppf, a=0)
        # import matplotlib.pyplot as pt
        # pt.hist(rad, bins=30)
        # print(rad)
        # pt.show()

        # TODO test with different models

    # TODO rework this
    # def test_gau(self):
    #    for d in range(len(self.rngs)):
    #        Z, k = self.rngs[d]('gau', self.len_scale, self.many_modes)
    #        self.assertEqual(k.shape, (d+1, self.many_modes))
    #        self.assertAlmostEqual(np.mean(k), 0., places=2)
    #        self.assertAlmostEqual(np.std(k), 1/self.len_scale, places=2)
    #        self.assertAlmostEqual(skew(k[0, :]), 0., places=2)
    #        self.assertAlmostEqual(kurtosis(k[0, :]), 0., places=1)
    #        self.assertLess(normaltest(k[0, :])[1], 0.05)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
GStools subpackage providing tools for the spatial random field.

.. currentmodule:: gstools.field.tools

The following classes and functions are provided

.. autosummary::
   reshape_input
   reshape_input_axis_from_unstruct
   reshape_input_axis_from_struct
   check_mesh
   make_isotropic
   unrotate_mesh
   reshape_axis_from_struct_to_unstruct
   reshape_field_from_unstruct_to_struct
"""
# pylint: disable=C0103
from __future__ import print_function, division, absolute_import

import numpy as np
from gstools.tools.geometric import r3d_x, r3d_y, r3d_z

__all__ = [
    "reshape_input",
    "reshape_input_axis_from_unstruct",
    "reshape_input_axis_from_struct",
    "check_mesh",
    "make_isotropic",
    "unrotate_mesh",
    "reshape_axis_from_struct_to_unstruct",
    "reshape_field_from_unstruct_to_struct",
]


# Geometric functions #########################################################


def reshape_input(x, y=None, z=None, mesh_type="unstructured"):
    """Reshape given axes, depending on the mesh type."""
    if mesh_type == "unstructured":
        x, y, z = reshape_input_axis_from_unstruct(x, y, z)
    elif mesh_type == "structured":
        x, y, z = reshape_input_axis_from_struct(x, y, z)
    return x, y, z


def reshape_input_axis_from_unstruct(x, y=None, z=None):
    """Reshape given axes for vectorisation on unstructured grid."""
    x = np.atleast_1d(x)
    y = np.atleast_1d(y)
    z = np.atleast_1d(z)
    x = np.reshape(x, (len(x), 1))
    y = np.reshape(y, (len(y), 1))
    z = np.reshape(z, (len(z), 1))
    return (x, y, z)


def reshape_input_axis_from_struct(x, y=None, z=None):
    """Reshape given axes for vectorisation on unstructured grid."""
    x = np.atleast_1d(x)
    y = np.atleast_1d(y)
    z = np.atleast_1d(z)
    x = np.reshape(x, (len(x), 1, 1, 1))
    y = np.reshape(y, (1, len(y), 1, 1))
    z = np.reshape(z, (1, 1, len(z), 1))
    return (x, y, z)


# SRF helpers #################################################################


def check_mesh(dim, x, y, z, mesh_type):
    """Do a basic check of the shapes of the input arrays."""
    if dim >= 2:
        if y is None:
            raise ValueError(
                "The y-component is missing for " "{0} dimensions".format(dim)
            )
    if dim == 3:
        if z is None:
            raise ValueError(
                "The z-component is missing for " "{0} dimensions".format(dim)
            )
    if mesh_type == "unstructured":
        if dim >= 2:
            try:
                if len(x) != len(y):
                    raise ValueError(
                        "len(x) = {0} != len(y) = {1} "
                        "for unstructured grids".format(len(x), len(y))
                    )
            except TypeError:
                pass
            if dim == 3:
                try:
                    if len(x) != len(z):
                        raise ValueError(
                            "len(x) = {0} != len(z) = {1} "
                            "for unstructured grids".format(len(x), len(z))
                        )
                except TypeError:
                    pass
    elif mesh_type == "structured":
        pass
    else:
        raise ValueError("Unknown mesh type {0}".format(mesh_type))


def make_isotropic(dim, anis, y, z):
    """Stretch given axes in order to implement anisotropy."""
    if dim == 1:
        return y, z
    if dim == 2:
        return y / anis[0], z
    if dim == 3:
        return y / anis[0], z / anis[1]
    return None


def unrotate_mesh(dim, angles, x, y, z):
    """Rotate axes in order to implement rotation.

    for 3d: yaw, pitch, and roll angles are alpha, beta, and gamma,
    of intrinsic rotation rotation whose Tait-Bryan angles are
    alpha, beta, gamma about axes x, y, z.
    """
    if dim == 1:
        return x, y, z
    if dim == 2:
        # extract 2d rotation matrix
        rot_mat = r3d_z(-angles[0])[0:2, 0:2]
        pos_tuple = np.vstack((x, y))
        pos_tuple = np.vsplit(np.dot(rot_mat, pos_tuple), 2)
        x = pos_tuple[0].reshape(np.shape(x))
        y = pos_tuple[1].reshape(np.shape(y))
        return x, y, z
    if dim == 3:
        alpha = -angles[0]
        beta = -angles[1]
        gamma = -angles[2]
        rot_mat = np.dot(np.dot(r3d_z(alpha), r3d_y(beta)), r3d_x(gamma))
        pos_tuple = np.vstack((x, y, z))
        pos_tuple = np.vsplit(np.dot(rot_mat, pos_tuple), 3)
        x = pos_tuple[0].reshape(np.shape(x))
        y = pos_tuple[1].reshape(np.shape(y))
        z = pos_tuple[2].reshape(np.shape(z))
        return x, y, z
    return None


def reshape_axis_from_struct_to_unstruct(
    dim, x, y=None, z=None, indexing="ij"
):
    """Reshape given axes from struct to unstruct for rotation."""
    if dim == 1:
        return x, y, z, (len(x),)
    if dim == 2:
        x_u, y_u = np.meshgrid(x, y, indexing=indexing)
        len_unstruct = len(x) * len(y)
        x_u = np.reshape(x_u, len_unstruct)
        y_u = np.reshape(y_u, len_unstruct)
        return x_u, y_u, z, (len(x), len(y))
    if dim == 3:
        x_u, y_u, z_u = np.meshgrid(x, y, z, indexing=indexing)
        len_unstruct = len(x) * len(y) * len(z)
        x_u = np.reshape(x_u, len_unstruct)
        y_u = np.reshape(y_u, len_unstruct)
        z_u = np.reshape(z_u, len_unstruct)
        return x_u, y_u, z_u, (len(x), len(y), len(z))
    return None


def reshape_field_from_unstruct_to_struct(dim, field, axis_lens):
    """Reshape the rotated field back to struct."""
    if dim == 1:
        return field
    if dim == 2:
        field = np.reshape(field, axis_lens)
        return field
    if dim == 3:
        field = np.reshape(field, axis_lens)
        return field
    return None


def _get_select(direction):
    select = []
    if not (0 < len(direction) < 4):
        raise ValueError(
            "Field.mesh: need 1 to 3 direction(s), got '{}'".format(direction)
        )
    for axis in direction:
        if axis == "x":
            if 0 in select:
                raise ValueError(
                    "Field.mesh: got duplicate directions {}".format(direction)
                )
            select.append(0)
        elif axis == "y":
            if 1 in select:
                raise ValueError(
                    "Field.mesh: got duplicate directions {}".format(direction)
                )
            select.append(1)
        elif axis == "z":
            if 2 in select:
                raise ValueError(
                    "Field.mesh: got duplicate directions {}".format(direction)
                )
            select.append(2)
        else:
            raise ValueError(
                "Field.mesh: got unknown direction {}".format(axis)
            )
    return select



# -*- coding: utf-8 -*-
"""
GStools subpackage providing tools for spatial random fields.

.. currentmodule:: gstools.field

Subpackages
^^^^^^^^^^^

.. autosummary::
    generator
    upscaling
    base

Spatial Random Field
^^^^^^^^^^^^^^^^^^^^

.. autosummary::
   SRF

----
"""
from __future__ import absolute_import

from gstools.field.srf import SRF

__all__ = ["SRF"]



# -*- coding: utf-8 -*-
"""
GStools subpackage for random number generation.

.. currentmodule:: gstools.random

Random Number Generator
^^^^^^^^^^^^^^^^^^^^^^^

.. autosummary::
   RNG

Seed Generator
^^^^^^^^^^^^^^

.. autosummary::
    MasterRNG

Distribution factory
^^^^^^^^^^^^^^^^^^^^

.. autosummary::
   dist_gen

----
"""
from __future__ import absolute_import

from gstools.random.rng import RNG
from gstools.random.tools import MasterRNG, dist_gen

__all__ = ["RNG", "MasterRNG", "dist_gen"]



# -*- coding: utf-8 -*-
"""
GStools subpackage providing a set of handy covariance models.

Subpackages
^^^^^^^^^^^

.. currentmodule:: gstools.covmodel

.. autosummary::
    base
    models
    tpl_models
    plot

Covariance Base-Class
^^^^^^^^^^^^^^^^^^^^^
Class to construct user defined covariance models

.. currentmodule:: gstools.covmodel.base

.. autosummary::
   CovModel

Covariance Models
^^^^^^^^^^^^^^^^^
Standard Covariance Models

.. currentmodule:: gstools.covmodel.models

.. autosummary::
   Gaussian
   Exponential
   Matern
   Rational
   Stable
   Linear
   Circular
   Spherical
   Intersection

Truncated Power Law Covariance Models

.. currentmodule:: gstools.covmodel.tpl_models

.. autosummary::
   TPLGaussian
   TPLExponential
   TPLStable
"""
from __future__ import absolute_import

from gstools.covmodel.base import CovModel
from gstools.covmodel.models import (
    Gaussian,
    Exponential,
    Matern,
    Rational,
    Stable,
    Linear,
    Circular,
    Spherical,
    Intersection,
)
from gstools.covmodel.tpl_models import TPLGaussian, TPLExponential, TPLStable

__all__ = [
    "CovModel",
    "Gaussian",
    "Exponential",
    "Matern",
    "Rational",
    "Stable",
    "Linear",
    "Circular",
    "Spherical",
    "Intersection",
    "TPLGaussian",
    "TPLExponential",
    "TPLStable",
]



# -*- coding: utf-8 -*-
"""GSTools: A geostatistical toolbox."""
from __future__ import division, absolute_import, print_function
import sys, os, codecs, re, tempfile, glob, subprocess, shutil

from distutils.errors import CompileError, LinkError
from distutils.ccompiler import new_compiler
from distutils.sysconfig import customize_compiler

from setuptools import setup, find_packages, Distribution, Extension
from Cython.Build import cythonize
import numpy


HERE = os.path.abspath(os.path.dirname(__file__))


# version finder ##############################################################


def read(*file_paths):
    """Read file data."""
    with codecs.open(os.path.join(HERE, *file_paths), "r") as file_in:
        return file_in.read()


def find_version(*file_paths):
    """Find version without importing module."""
    version_file = read(*file_paths)
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M
    )
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


# openmp finder ###############################################################
# This code is adapted for a large part from the scikit-learn openmp helpers,
# which can be found at:
# https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/_build_utils/openmp_helpers.py


# TemporaryDirectory not avialable in python2
class _TemporaryDirectory(object):
    def __enter__(self):
        self.dir_name = tempfile.mkdtemp()
        return self.dir_name

    def __exit__(self, exc_type, exc_value, traceback):
        shutil.rmtree(self.dir_name)


TemporaryDirectory = getattr(
    tempfile, "TemporaryDirectory", _TemporaryDirectory
)

CCODE = """
#include <omp.h>
#include <stdio.h>
int main(void) {
#pragma omp parallel
printf("nthreads=%d\\n", omp_get_num_threads());
return 0;
}
"""


def get_openmp_flag(compiler):
    """Get the compiler dependent openmp flag."""
    if hasattr(compiler, "compiler"):
        compiler = compiler.compiler[0]
    else:
        compiler = compiler.__class__.__name__

    if sys.platform == "win32" and ("icc" in compiler or "icl" in compiler):
        return ["/Qopenmp"]
    if sys.platform == "win32":
        return ["/openmp"]
    if sys.platform == "darwin" and ("icc" in compiler or "icl" in compiler):
        return ["-openmp"]
    if sys.platform == "darwin" and "openmp" in os.getenv("CPPFLAGS", ""):
        # -fopenmp can't be passed as compile flag when using Apple-clang.
        # OpenMP support has to be enabled during preprocessing.
        #
        # For example, our macOS wheel build jobs use the following environment
        # variables to build with Apple-clang and the brew installed "libomp":
        #
        # export CPPFLAGS="$CPPFLAGS -Xpreprocessor -fopenmp"
        # export CFLAGS="$CFLAGS -I/usr/local/opt/libomp/include"
        # export CXXFLAGS="$CXXFLAGS -I/usr/local/opt/libomp/include"
        # export LDFLAGS="$LDFLAGS -L/usr/local/opt/libomp/lib -lomp"
        # export DYLD_LIBRARY_PATH=/usr/local/opt/libomp/lib
        return []
    # Default flag for GCC and clang:
    return ["-fopenmp"]


def check_openmp_support():
    """Check whether OpenMP test code can be compiled and run."""
    ccompiler = new_compiler()
    customize_compiler(ccompiler)

    with TemporaryDirectory() as tmp_dir:
        try:
            os.chdir(tmp_dir)
            # Write test program
            with open("test_openmp.c", "w") as cfile:
                cfile.write(CCODE)
            os.mkdir("objects")
            # Compile, test program
            openmp_flags = get_openmp_flag(ccompiler)
            ccompiler.compile(
                ["test_openmp.c"],
                output_dir="objects",
                extra_postargs=openmp_flags,
            )
            # Link test program
            extra_preargs = os.getenv("LDFLAGS", None)
            if extra_preargs is not None:
                extra_preargs = extra_preargs.split(" ")
            else:
                extra_preargs = []
            objects = glob.glob(
                os.path.join("objects", "*" + ccompiler.obj_extension)
            )
            ccompiler.link_executable(
                objects,
                "test_openmp",
                extra_preargs=extra_preargs,
                extra_postargs=openmp_flags,
            )
            # Run test program
            output = subprocess.check_output("./test_openmp")
            output = output.decode(sys.stdout.encoding or "utf-8").splitlines()
            # Check test program output
            if "nthreads=" in output[0]:
                nthreads = int(output[0].strip().split("=")[1])
                openmp_supported = len(output) == nthreads
            else:
                openmp_supported = False
                openmp_flags = []
        except (CompileError, LinkError, subprocess.CalledProcessError):
            openmp_supported = False
            openmp_flags = []
        finally:
            os.chdir(HERE)
    return openmp_supported, openmp_flags


# openmp ######################################################################


USE_OPENMP = bool("--openmp" in sys.argv)

if USE_OPENMP:
    # just check if wanted
    CAN_USE_OPENMP, FLAGS = check_openmp_support()
    if CAN_USE_OPENMP:
        print("## GSTOOLS setup: OpenMP found.")
        print("## OpenMP flags:", FLAGS)
    else:
        print("## GSTOOLS setup: OpenMP not found.")
else:
    CAN_USE_OPENMP = False
    print("## GSTOOLS setup: OpenMP not wanted by the user.")
    FLAGS = []

USE_OPENMP = USE_OPENMP and CAN_USE_OPENMP


# add the "--openmp" to the global options
# enables calles like:
# python3 setup.py --openmp build_ext --inplace
# pip install --global-option="--openmp" gstools
class MPDistribution(Distribution):
    """Distribution with --openmp as global option."""

    global_options = Distribution.global_options + [
        ("openmp", None, "Flag to use openmp in the build")
    ]


# cython extensions ###########################################################


EXT_MODULES = []

EXTRA_COMPILE_ARGS = FLAGS
EXTRA_LINK_ARGS = FLAGS

SUMMATOR_EXT = Extension(
    "gstools.field.summator",
    [os.path.join("gstools", "field", "summator.pyx")],
    include_dirs=[numpy.get_include()],
    extra_compile_args=EXTRA_COMPILE_ARGS,
    extra_link_args=EXTRA_LINK_ARGS,
)
VARIOGRAM_EXT = Extension(
    "gstools.variogram.estimator",
    [os.path.join("gstools", "variogram", "estimator.pyx")],
    include_dirs=[numpy.get_include()],
    extra_compile_args=EXTRA_COMPILE_ARGS,
    extra_link_args=EXTRA_LINK_ARGS,
)

KRIGESUM_EXT = Extension(
    "gstools.krige.krigesum",
    [os.path.join("gstools", "krige", "krigesum.pyx")],
    include_dirs=[numpy.get_include()],
    extra_compile_args=EXTRA_COMPILE_ARGS,
    extra_link_args=EXTRA_LINK_ARGS,
)

EXT_MODULES += cythonize(
    [VARIOGRAM_EXT, SUMMATOR_EXT, KRIGESUM_EXT],
    # annotate=True
)

# This is the important part. By setting this compiler directive, cython will
# embed signature information in docstrings. Sphinx then knows how to extract
# and use those signatures.
# python setup.py build_ext --inplace --> then sphinx build
for ext_m in EXT_MODULES:
    ext_m.cython_directives = {"embedsignature": True}


# setup #######################################################################


# version import not possible due to cython
# see: https://packaging.python.org/guides/single-sourcing-package-version/
VERSION = find_version("gstools", "_version.py")
DOCLINES = __doc__.split("\n")
README = read("README.md")

CLASSIFIERS = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "Intended Audience :: End Users/Desktop",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
    "Natural Language :: English",
    "Operating System :: Unix",
    "Programming Language :: Python",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 3",
    "Topic :: Scientific/Engineering",
    "Topic :: Utilities",
]

setup(
    name="gstools",
    version=VERSION,
    maintainer="Lennart Schueler, Sebastian Mueller",
    maintainer_email="info@geostat-framework.org",
    description=DOCLINES[0],
    long_description=README,
    long_description_content_type="text/markdown",
    author="Lennart Schueler, Sebastian Mueller",
    author_email="info@geostat-framework.org",
    url="https://github.com/GeoStat-Framework/GSTools",
    license="LGPLv3",
    classifiers=CLASSIFIERS,
    platforms=["Windows", "Linux", "Solaris", "Mac OS-X", "Unix"],
    include_package_data=True,
    setup_requires=[
        "numpy>=1.14.5",  # numpy imported in setup.py
        "cython>=0.28.3",
        "setuptools>=41.0.1",
    ],
    install_requires=[
        "numpy>=1.14.5",
        "scipy>=1.1.0",
        "hankel>=0.3.6",
        "emcee>=3.0.0",
        "pyevtk",
        "six",
    ],
    extras_require={"plotting": ["pyvista", "matplotlib"]},
    packages=find_packages(exclude=["tests*", "docs*"]),
    ext_modules=EXT_MODULES,
    include_dirs=[numpy.get_include()],
    distclass=MPDistribution,
)
